# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import http
import re
import typing
import warnings
from os import environ
from typing import Callable, Optional, Union

import requests.exceptions
from nuclio.build import mlrun_footer

import mlrun.common.constants
import mlrun.common.constants as mlrun_constants
import mlrun.common.formatters
import mlrun.common.runtimes
import mlrun.common.schemas
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.errors
import mlrun.launcher.factory
import mlrun.utils.helpers
import mlrun.utils.notifications
import mlrun.utils.regex
from mlrun.common.runtimes.constants import RunStates
from mlrun.model import (
    BaseMetadata,
    HyperParamOptions,
    ImageBuilder,
    ModelObj,
    RunObject,
    RunTemplate,
)
from mlrun.utils.helpers import generate_object_uri, verify_field_regex
from mlrun_pipelines.common.ops import mlrun_op

from ..config import config
from ..datastore import store_manager
from ..errors import err_to_str
from ..lists import RunList
from ..utils import (
    dict_to_json,
    dict_to_yaml,
    enrich_image_url,
    get_in,
    get_parsed_docker_registry,
    logger,
    now_date,
    update_in,
)
from .funcdoc import update_function_entry_points
from .utils import RunError, calc_hash

spec_fields = [
    "command",
    "args",
    "image",
    "mode",
    "build",
    "entry_points",
    "description",
    "workdir",
    "default_handler",
    "pythonpath",
    "disable_auto_mount",
    "allow_empty_resources",
    "reset_on_run",
]


class RuntimeClassMode(enum.Enum):
    """
    Runtime class mode
    Currently there are two modes:
    * run - the runtime class is used to run a function
    * build - the runtime class is used to build a function

    The runtime class mode is used to determine what should be the name of the runtime class, each runtime might have a
    different name for each mode and some might not have both modes.
    """

    run = "run"
    build = "build"


class FunctionStatus(ModelObj):
    def __init__(self, state=None, build_pod=None):
        self.state = state
        self.build_pod = build_pod


class FunctionSpec(ModelObj):
    _dict_fields = spec_fields
    _default_fields_to_strip = []

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        build=None,
        entry_points=None,
        description=None,
        workdir=None,
        default_handler=None,
        pythonpath=None,
        disable_auto_mount=False,
    ):
        self.command = command or ""
        self.image = image or ""
        self.mode = mode
        self.args = args or []
        self.rundb = None
        self.description = description or ""
        self.workdir = workdir
        self.pythonpath = pythonpath

        self._build = None
        self.build = build
        self.default_handler = default_handler
        self.entry_points = entry_points or {}
        self.disable_auto_mount = disable_auto_mount
        self.allow_empty_resources = None

    @property
    def build(self) -> ImageBuilder:
        return self._build

    @build.setter
    def build(self, build):
        self._build = self._verify_dict(build, "build", ImageBuilder)

    def validate_service_account(self, allowed_service_accounts):
        pass


class BaseRuntime(ModelObj):
    kind = "base"
    _is_nested = False
    _is_remote = False
    _dict_fields = ["kind", "metadata", "spec", "status", "verbose"]
    _default_fields_to_strip = ModelObj._default_fields_to_strip + [
        "status",  # Function status describes the state rather than configuration
    ]

    def __init__(self, metadata=None, spec=None):
        self._metadata = None
        self.metadata = metadata
        self.kfp = None
        self._spec = None
        self.spec = spec
        self._db_conn = None
        self._secrets = None
        self._k8s = None
        self._is_built = False
        self.is_child = False
        self._status = None
        self.status = None
        self.verbose = False
        self._enriched_image = False

    def set_db_connection(self, conn):
        if not self._db_conn:
            self._db_conn = conn

    @property
    def metadata(self) -> BaseMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = self._verify_dict(metadata, "metadata", BaseMetadata)

    @property
    def spec(self) -> FunctionSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", FunctionSpec)

    @property
    def status(self) -> FunctionStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", FunctionStatus)

    def set_label(self, key, value):
        self.metadata.labels[key] = str(value)
        return self

    def set_categories(self, categories: list[str]):
        self.metadata.categories = mlrun.utils.helpers.as_list(categories)

    @property
    def uri(self):
        return self._function_uri()

    def is_deployed(self):
        return True

    def is_model_monitoring_function(self):
        return (
            self.metadata.labels.get(mm_constants.ModelMonitoringAppLabel.KEY, "")
            == mm_constants.ModelMonitoringAppLabel.VAL
        )

    def _is_remote_api(self):
        db = self._get_db()
        if db and db.kind == "http":
            return True
        return False

    def _function_uri(self, tag=None, hash_key=None):
        return generate_object_uri(
            self.metadata.project,
            self.metadata.name,
            tag=tag or self.metadata.tag,
            hash_key=hash_key,
        )

    def _ensure_run_db(self):
        self.spec.rundb = self.spec.rundb or mlrun.db.get_or_set_dburl()

    def _get_db(self):
        # TODO: remove this function and use the launcher db instead
        self._ensure_run_db()
        if not self._db_conn:
            if self.spec.rundb:
                self._db_conn = mlrun.db.get_run_db(
                    self.spec.rundb, secrets=self._secrets
                )
        return self._db_conn

    # This function is different than the auto_mount function, as it mounts to runtimes based on the configuration.
    # That's why it's named differently.
    def try_auto_mount_based_on_config(self):
        pass

    def validate_and_enrich_service_account(
        self, allowed_service_account, default_service_account
    ):
        pass

    def _fill_credentials(self):
        """
        If access key is not mask (starts with secret prefix) then fill $generate so that the API will handle filling
         of the credentials.
        We rely on the HTTPDB to send the access key session through the request header and that the API will mask
         the access key, that way we won't even store any plain access key in the function.
        """
        if self.metadata.credentials.access_key and (
            # if contains secret reference or $generate then no need to overwrite the access key
            self.metadata.credentials.access_key.startswith(
                mlrun.model.Credentials.secret_reference_prefix
            )
            or self.metadata.credentials.access_key.startswith(
                mlrun.model.Credentials.generate_access_key
            )
        ):
            return
        self.metadata.credentials.access_key = (
            mlrun.model.Credentials.generate_access_key
        )

    def generate_runtime_k8s_env(self, runobj: RunObject = None) -> list[dict]:
        """
        Prepares a runtime environment as it's expected by kubernetes.models.V1Container

        :param runobj: Run context object (RunObject) with run metadata and status
        :return: List of dicts with the structure {"name": "var_name", "value": "var_value"}
        """
        return [
            {"name": k, "value": v}
            for k, v in self._generate_runtime_env(runobj).items()
        ]

    def run(
        self,
        runspec: Optional[
            Union["mlrun.run.RunTemplate", "mlrun.run.RunObject", dict]
        ] = None,
        handler: Optional[Union[str, Callable]] = None,
        name: Optional[str] = "",
        project: Optional[str] = "",
        params: Optional[dict] = None,
        inputs: Optional[dict[str, str]] = None,
        out_path: Optional[str] = "",
        workdir: Optional[str] = "",
        artifact_path: Optional[str] = "",
        watch: Optional[bool] = True,
        schedule: Optional[Union[str, mlrun.common.schemas.ScheduleCronTrigger]] = None,
        hyperparams: Optional[dict[str, list]] = None,
        hyper_param_options: Optional[HyperParamOptions] = None,
        verbose: Optional[bool] = None,
        scrape_metrics: Optional[bool] = None,
        local: Optional[bool] = False,
        local_code_path: Optional[str] = None,
        auto_build: Optional[bool] = None,
        param_file_secrets: Optional[dict[str, str]] = None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        reset_on_run: Optional[bool] = None,
        output_path: Optional[str] = "",
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
        **launcher_kwargs,
    ) -> RunObject:
        """
        Run a local or remote task.

        :param runspec:        The run spec to generate the RunObject from. Can be RunTemplate | RunObject | dict.
        :param handler:        Pointer or name of a function handler.
        :param name:           Execution name.
        :param project:        Project name.
        :param params:         Input parameters (dict).
        :param inputs:         Input objects to pass to the handler. Type hints can be given so the input will be parsed
                               during runtime from `mlrun.DataItem` to the given type hint. The type hint can be given
                               in the key field of the dictionary after a colon, e.g: "<key> : <type_hint>".
        :param out_path:       (deprecated) Default artifact output path.
        :param artifact_path:  (deprecated) Default artifact output path (will replace out_path).
        :param workdir:        Working directory of the executed job and the default path for artifact inputs
        :param watch:          Watch/follow run log.
        :param schedule:       ScheduleCronTrigger class instance or a standard crontab expression string
                               (which will be converted to the class using its `from_crontab` constructor),
                               see this link for help:
                               https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html#module-apscheduler.triggers.cron
        :param hyperparams:    Dict of param name and list of values to be enumerated.
                               The default strategy is grid search and uses e.g. {"p1": [1,2,3]}.
                               (Can be specified as a JSON file)
                               For list, lists must be of equal length, e.g. {"p1": [1], "p2": [2]}.
                               (Can be specified as JSON file or as a CSV file listing the parameter values
                               per iteration.)
                               You can specify strategy of type grid, list, random,
                               and other options in the hyper_param_options parameter.
        :param hyper_param_options: Dict or :py:class:`~mlrun.model.HyperParamOptions` struct of hyperparameter options.
        :param verbose:             Add verbose prints/logs.
        :param scrape_metrics:      Whether to add the `mlrun/scrape-metrics` label to this run's resources.
        :param local:               Run the function locally vs on the runtime/cluster.
        :param local_code_path:     Path of the code for local runs & debug.
        :param auto_build:          When set to True and the function require build it will be built on the first
                                    function run, use only if you don't plan on changing the build config between runs.
        :param param_file_secrets:  Dictionary of secrets to be used only for accessing the hyper-param parameter file.
                                    These secrets are only used locally and will not be stored anywhere
        :param notifications:       List of notifications to push when the run is completed
        :param returns: List of log hints - configurations for how to log the returning values from the handler's run
                        (as artifacts or results). The list's length must be equal to the amount of returning objects. A
                        log hint may be given as:

                        * A string of the key to use to log the returning value as result or as an artifact. To specify
                          The artifact type, it is possible to pass a string in the following structure:
                          "<key> : <type>". Available artifact types can be seen in `mlrun.ArtifactType`. If no
                          artifact type is specified, the object's default artifact type will be used.
                        * A dictionary of configurations to use when logging. Further info per object type and artifact
                          type can be given there. The artifact key must appear in the dictionary as "key": "the_key".
        :param state_thresholds:    Dictionary of states to time thresholds. The state will be matched against the
                k8s resource's status. The threshold should be a time string that conforms to timelength python package
                standards and is at least 1 minute (-1 for infinite).
                If the phase is active for longer than the threshold, the run will be aborted.
                See mlconf.function.spec.state_thresholds for the state options and default values.
        :param reset_on_run: When True, function python modules would reload prior to code execution.
                             This ensures latest code changes are executed. This argument must be used in
                             conjunction with the local=True argument.
        :param output_path:    Default artifact output path.
        :param retry:          Retry configuration for the run, can be a dict or an instance of
                               :py:class:`~mlrun.model.Retry`.
                               The `count` field in the `Retry` object specifies the number of retry attempts.
                               If `count=0`, the run will not be retried.
                               The `backoff` field specifies the retry backoff strategy between retry attempts.
                               If not provided, the default backoff delay is 30 seconds.
        :return: Run context object (RunObject) with run metadata, results and status
        """
        if artifact_path or out_path:
            deprecated_param = "artifact_path" if artifact_path else "out_path"
            warnings.warn(
                f"'{deprecated_param}' parameter is deprecated in 1.10.0 and will be removed in 1.12.0, "
                "use 'output_path' instead.",
                # TODO: Remove this in 1.12.0
                FutureWarning,
            )
        output_path = output_path or out_path or artifact_path

        mlrun.utils.helpers.validate_function_name(self.metadata.name)

        launcher = mlrun.launcher.factory.LauncherFactory().create_launcher(
            self._is_remote, local=local, **launcher_kwargs
        )
        return launcher.launch(
            runtime=self,
            task=runspec,
            handler=handler,
            name=name,
            project=project,
            params=params,
            inputs=inputs,
            workdir=workdir,
            output_path=output_path,
            watch=watch,
            schedule=schedule,
            hyperparams=hyperparams,
            hyper_param_options=hyper_param_options,
            verbose=verbose,
            scrape_metrics=scrape_metrics,
            local_code_path=local_code_path,
            auto_build=auto_build,
            param_file_secrets=param_file_secrets,
            notifications=notifications,
            returns=returns,
            state_thresholds=state_thresholds,
            reset_on_run=reset_on_run,
            retry=retry,
        )

    def _get_db_run(
        self,
        task: RunObject = None,
        run_format: mlrun.common.formatters.RunFormat = mlrun.common.formatters.RunFormat.full,
    ):
        if self._get_db() and task:
            project = task.metadata.project
            uid = task.metadata.uid
            iter = task.metadata.iteration
            try:
                return self._get_db().read_run(
                    uid, project, iter=iter, format_=run_format
                )
            except mlrun.db.RunDBError:
                return None
        if task:
            return task.to_dict()

    def _generate_runtime_env(self, runobj: RunObject = None) -> dict:
        """
        Prepares all available environment variables for usage on a runtime
        Data will be extracted from several sources and most of them are not guaranteed to be available

        :param runobj: Run context object (RunObject) with run metadata and status
        :return: Dictionary with all the variables that could be parsed
        """
        active_project = self.metadata.project or config.active_project
        runtime_env = {
            mlrun_constants.MLRUN_ACTIVE_PROJECT: active_project,
            # TODO: Remove this in 1.12.0 as MLRUN_DEFAULT_PROJECT is deprecated and should not be injected anymore
            "MLRUN_DEFAULT_PROJECT": active_project,
        }
        if runobj:
            runtime_env["MLRUN_EXEC_CONFIG"] = runobj.to_json(
                exclude_notifications_params=True
            )
            if runobj.metadata.project:
                runtime_env[mlrun_constants.MLRUN_ACTIVE_PROJECT] = (
                    runobj.metadata.project
                )
            if runobj.spec.verbose:
                runtime_env["MLRUN_LOG_LEVEL"] = "DEBUG"
        if config.httpdb.api_url:
            runtime_env["MLRUN_DBPATH"] = config.httpdb.api_url
        if self.metadata.namespace or config.namespace:
            runtime_env["MLRUN_NAMESPACE"] = self.metadata.namespace or config.namespace
        return runtime_env

    @staticmethod
    def _handle_submit_job_http_error(error: requests.HTTPError):
        # if we receive a 400 status code, this means the request was invalid and the run wasn't created in the DB.
        # so we don't need to update the run state and we can just raise the error.
        # more status code handling can be added here if needed
        if error.response.status_code == http.HTTPStatus.BAD_REQUEST.value:
            raise mlrun.errors.MLRunBadRequestError(
                f"Bad request to mlrun api: {error.response.text}"
            )

    def _store_function(self, runspec, meta, db):
        meta.labels["kind"] = self.kind
        mlrun.runtimes.utils.enrich_run_labels(
            meta.labels, [mlrun_constants.MLRunInternalLabels.owner]
        )
        if runspec.spec.output_path:
            runspec.spec.output_path = runspec.spec.output_path.replace(
                "{{run.user}}", meta.labels[mlrun_constants.MLRunInternalLabels.owner]
            )

        if db and self.kind != "handler":
            struct = self.to_dict()
            hash_key = db.store_function(
                struct, self.metadata.name, self.metadata.project, versioned=True
            )
            runspec.spec.function = self._function_uri(hash_key=hash_key)

    def _pre_run(self, runspec: RunObject, execution):
        pass

    def _post_run(self, results, execution):
        pass

    def _run(self, runobj: RunObject, execution) -> dict:
        pass

    def _run_many(self, generator, execution, runobj: RunObject) -> RunList:
        results = RunList()
        num_errors = 0
        tasks = generator.generate(runobj)
        for task in tasks:
            try:
                self.store_run(task)
                resp = self._run(task, execution)
                resp = self._update_run_state(resp, task=task)
                run_results = resp["status"].get("results", {})
                if generator.eval_stop_condition(run_results):
                    logger.info(
                        f"Reached early stop condition ({generator.options.stop_condition}), stopping iterations!"
                    )
                    results.append(resp)
                    break

            except RunError as err:
                task.status.state = "error"
                error_string = err_to_str(err)
                task.status.error = error_string
                resp = self._update_run_state(task=task, err=error_string)
                num_errors += 1
                if num_errors > generator.max_errors:
                    logger.error("Too many errors, stopping iterations!")
                    results.append(resp)
                    break

            results.append(resp)

        return results

    def store_run(self, runobj: RunObject):
        if self._get_db() and runobj:
            project = runobj.metadata.project
            uid = runobj.metadata.uid
            iter = runobj.metadata.iteration
            self._get_db().store_run(runobj.to_dict(), uid, project, iter=iter)

    def _store_run_dict(self, rundict: dict):
        if self._get_db() and rundict:
            project = get_in(rundict, "metadata.project", "")
            uid = get_in(rundict, "metadata.uid")
            iter = get_in(rundict, "metadata.iteration", 0)
            self._get_db().store_run(rundict, uid, project, iter=iter)

    def _update_run_state(
        self,
        resp: Optional[dict] = None,
        task: RunObject = None,
        err: Optional[Union[Exception, str]] = None,
        run_format: mlrun.common.formatters.RunFormat = mlrun.common.formatters.RunFormat.full,
    ) -> typing.Optional[dict]:
        """update the task state in the DB"""
        was_none = False
        if resp is None and task:
            was_none = True
            resp = self._get_db_run(task, run_format)

            if not resp:
                self.store_run(task)
                return task.to_dict()

            if task.status.status_text:
                update_in(resp, "status.status_text", task.status.status_text)

        if resp is None:
            return None

        if not isinstance(resp, dict):
            raise ValueError(f"post_run called with type {type(resp)}")

        updates = None
        last_state = get_in(resp, "status.state", "")
        kind = get_in(resp, "metadata.labels.kind", "")
        if last_state in RunStates.error_states() or err:
            new_state = RunStates.error
            status_text = None
            max_retries = get_in(resp, "spec.retry.count", 0)
            retry_count = get_in(resp, "status.retry_count", 0) or 0
            attempts = retry_count + 1
            if max_retries:
                if retry_count < max_retries:
                    new_state = RunStates.pending_retry
                    status_text = f"Run failed attempt {attempts} of {max_retries + 1}"
                elif retry_count >= max_retries:
                    status_text = f"Run failed after {attempts} attempts"

            updates = {
                "status.last_update": now_date().isoformat(),
                "status.state": new_state,
            }
            update_in(resp, "status.state", new_state)
            if status_text:
                updates["status.status_text"] = status_text
                update_in(resp, "status.status_text", status_text)
            if err:
                update_in(resp, "status.error", err_to_str(err))
            err = get_in(resp, "status.error")
            if err:
                updates["status.error"] = err_to_str(err)

        elif (
            not was_none
            and last_state != RunStates.completed
            and last_state not in RunStates.error_and_abortion_states()
        ):
            try:
                runtime_cls = mlrun.runtimes.get_runtime_class(kind)
                updates = runtime_cls._get_run_completion_updates(resp)
            except KeyError:
                updates = self._get_run_completion_updates(resp)

        uid = get_in(resp, "metadata.uid")
        logger.debug(
            "Run updates",
            name=get_in(resp, "metadata.name"),
            uid=uid,
            kind=kind,
            last_state=last_state,
            updates=updates,
        )
        if self._get_db() and updates:
            project = get_in(resp, "metadata.project")
            iter = get_in(resp, "metadata.iteration", 0)
            self._get_db().update_run(updates, uid, project, iter=iter)

        return resp

    def _force_handler(self, handler):
        if not handler:
            raise RunError(f"Handler must be provided for {self.kind} runtime")

    def _has_pipeline_param(self) -> bool:
        # check if the runtime has pipeline parameters
        # https://www.kubeflow.org/docs/components/pipelines/v1/sdk/parameters/
        matches = re.findall(mlrun.utils.regex.pipeline_param[0], self.to_json())
        return bool(matches)

    @staticmethod
    def _get_run_completion_updates(run: dict) -> dict:
        """
        Get the required updates for the run object when it's completed and update the run object state
        Override this if the run completion is not resolved by a single execution
        """
        updates = {
            "status.last_update": now_date().isoformat(),
            "status.state": "completed",
        }
        update_in(run, "status.state", "completed")
        return updates

    def full_image_path(
        self,
        image=None,
        client_version: Optional[str] = None,
        client_python_version: Optional[str] = None,
    ):
        image = image or self.spec.image or ""

        image = enrich_image_url(image, client_version, client_python_version)
        if not image.startswith(
            mlrun.common.constants.IMAGE_NAME_ENRICH_REGISTRY_PREFIX
        ):
            return image
        registry, repository = get_parsed_docker_registry()
        if registry:
            if repository and repository not in image:
                return f"{registry}/{repository}/{image[1:]}"
            return f"{registry}/{image[1:]}"
        namespace_domain = environ.get("IGZ_NAMESPACE_DOMAIN", None)
        if namespace_domain is not None:
            return f"docker-registry.{namespace_domain}:80/{image[1:]}"
        raise RunError("Local container registry is not defined")

    def as_step(
        self,
        runspec: Union[RunObject, RunTemplate] = None,
        handler=None,
        name: str = "",
        project: str = "",
        params: Optional[dict] = None,
        hyperparams=None,
        selector="",
        hyper_param_options: HyperParamOptions = None,
        inputs: Optional[dict] = None,
        outputs: Optional[list] = None,
        workdir: str = "",
        artifact_path: str = "",
        image: str = "",
        labels: Optional[dict] = None,
        use_db=True,
        verbose=None,
        scrape_metrics=False,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        auto_build: bool = False,
    ):
        """Run a local or remote task.

        :param runspec:         run template object or dict (see RunTemplate)
        :param handler:         name of the function handler
        :param name:            execution name
        :param project:         project name
        :param params:          input parameters (dict)
        :param hyperparams:     hyper parameters
        :param selector:        selection criteria for hyper params
        :param hyper_param_options:  hyper param options (selector, early stop, strategy, ..)
                            see: :py:class:`~mlrun.model.HyperParamOptions`
        :param inputs:          Input objects to pass to the handler. Type hints can be given so the input will be
                                parsed during runtime from `mlrun.DataItem` to the given type hint. The type hint can be
                                given in the key field of the dictionary after a colon, e.g: "<key> : <type_hint>".
        :param outputs:         list of outputs which can pass in the workflow
        :param artifact_path:   default artifact output path (replace out_path)
        :param workdir:         working directory of the executed job and the default path for artifact inputs
        :param image:           container image to use
        :param labels:          labels to tag the job/run with ({key:val, ..})
        :param use_db:          save function spec in the db (vs the workflow file)
        :param verbose:         add verbose prints/logs
        :param scrape_metrics:  whether to add the `mlrun/scrape-metrics` label to this run's resources
        :param returns:         List of configurations for how to log the returning values from the handler's run
                                (as artifacts or results). The list's length must be equal to the amount of returning
                                objects. A configuration may be given as:

                                * A string of the key to use to log the returning value as result or as an artifact.
                                  To specify The artifact type, it is possible to pass a string in the following
                                  structure:
                                  "<key> : <type>". Available artifact types can be seen in `mlrun.ArtifactType`. If no
                                  artifact type is specified, the object's default artifact type will be used.
                                * A dictionary of configurations to use when logging. Further info per object type and
                                  artifact type can be given there. The artifact key must appear in the dictionary as
                                  "key": "the_key".
        :param auto_build:      when set to True and the function require build it will be built on the first
                                function run, use only if you dont plan on changing the build config between runs
        :return: mlrun_pipelines.models.PipelineNodeWrapper
        """

        # if the function contain KFP PipelineParams (futures) pass the full spec to the
        # PipelineNodeWrapper this way KFP will substitute the params with previous step outputs
        if use_db and not self._has_pipeline_param():
            # if the same function is built as part of the pipeline we do not use the versioned function
            # rather the latest function w the same tag so we can pick up the updated image/status
            versioned = False if hasattr(self, "_build_in_pipeline") else True
            url = self.save(versioned=versioned, refresh=True)
        else:
            url = None

        if runspec is not None:
            verify_field_regex(
                "run.metadata.name", runspec.metadata.name, mlrun.utils.regex.run_name
            )

        return mlrun_op(
            name,
            project,
            function=self,
            func_url=url,
            runobj=runspec,
            handler=handler,
            params=params,
            hyperparams=hyperparams,
            selector=selector,
            hyper_param_options=hyper_param_options,
            inputs=inputs,
            returns=returns,
            outputs=outputs,
            job_image=image,
            labels=labels,
            out_path=artifact_path,
            in_path=workdir,
            verbose=verbose,
            scrape_metrics=scrape_metrics,
            auto_build=auto_build,
        )

    def with_code(self, from_file="", body=None, with_doc=True):
        """Update the function code
        This function eliminates the need to build container images every time we edit the code

        :param from_file:   blank for current notebook, or path to .py/.ipynb file
        :param body:        will use the body as the function code
        :param with_doc:    update the document of the function parameters

        :return: function object
        """
        if body and from_file:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "must provide either body or from_file argument. not both"
            )

        if (not body and not from_file) or (from_file and from_file.endswith(".ipynb")):
            from nuclio import build_file

            _, _, body = build_file(from_file, name=self.metadata.name)

        else:
            if from_file:
                with open(from_file) as fp:
                    body = fp.read()
            if self.kind == mlrun.runtimes.RuntimeKinds.serving:
                body = body + mlrun_footer.format(
                    mlrun.runtimes.nuclio.serving.serving_subkind
                )

        self.spec.build.functionSourceCode = mlrun.utils.helpers.encode_user_code(body)
        if with_doc:
            update_function_entry_points(self, body)
        return self

    def with_requirements(
        self,
        requirements: Optional[list[str]] = None,
        overwrite: bool = False,
        prepare_image_for_deploy: bool = True,
        requirements_file: Optional[str] = "",
    ):
        """add package requirements from file or list to build spec.

        :param requirements:                a list of python packages
        :param requirements_file:           a local python requirements file path
        :param overwrite:                   overwrite existing requirements
        :param prepare_image_for_deploy:    prepare the image/base_image spec for deployment
        :return: function object
        """
        self.spec.build.with_requirements(requirements, requirements_file, overwrite)

        if prepare_image_for_deploy:
            self.prepare_image_for_deploy()

        return self

    def with_commands(
        self,
        commands: list[str],
        overwrite: bool = False,
        prepare_image_for_deploy: bool = True,
    ):
        """add commands to build spec.

        :param commands:                    list of commands to run during build
        :param overwrite:                   overwrite existing commands
        :param prepare_image_for_deploy:    prepare the image/base_image spec for deployment

        :return: function object
        """
        self.spec.build.with_commands(commands, overwrite)

        if prepare_image_for_deploy:
            self.prepare_image_for_deploy()
        return self

    def clean_build_params(self):
        # when using `with_requirements` we also execute `prepare_image_for_deploy` which adds the base image
        # and cleans the spec.image, so we need to restore the image back
        if self.spec.build.base_image and not self.spec.image:
            self.spec.image = self.spec.build.base_image

        self.spec.build = {}
        return self

    def requires_build(self) -> bool:
        build = self.spec.build
        return bool(
            build.commands
            or build.requirements
            or (build.source and not build.load_source_on_run)
        )

    def enrich_runtime_spec(
        self,
        project_node_selector: dict[str, str],
    ):
        pass

    def prepare_image_for_deploy(self):
        """
        if a function has a 'spec.image' it is considered to be deployed,
        but because we allow the user to set 'spec.image' for usability purposes,
        we need to check whether this is a built image or it requires to be built on top.
        """
        launcher = mlrun.launcher.factory.LauncherFactory().create_launcher(
            is_remote=self._is_remote
        )
        launcher.prepare_image_for_deploy(self)

    def export(self, target="", format=".yaml", secrets=None, strip=True):
        """save function spec to a local/remote path (default to./function.yaml)

        :param target:   target path/url
        :param format:   `.yaml` (default) or `.json`
        :param secrets:  optional secrets dict/object for target path (e.g. s3)
        :param strip:    strip status data

        :returns: self
        """
        if self.kind == "handler":
            raise ValueError(
                "cannot export local handler function, use "
                + "code_to_function() to serialize your function"
            )
        calc_hash(self)
        struct = self.to_dict(strip=strip)
        if format == ".yaml":
            data = dict_to_yaml(struct)
        else:
            data = dict_to_json(struct)
        stores = store_manager.set(secrets)
        target = target or "function.yaml"
        datastore, subpath, url = stores.get_or_create_store(target)
        datastore.put(subpath, data)
        logger.info(f"function spec saved to path: {target}")
        return self

    def save(self, tag="", versioned=False, refresh=False) -> str:
        launcher = mlrun.launcher.factory.LauncherFactory().create_launcher(
            is_remote=self._is_remote
        )
        return launcher.save_function(
            self, tag=tag, versioned=versioned, refresh=refresh
        )

    def doc(self):
        print("function:", self.metadata.name)
        print(self.spec.description)
        if self.spec.default_handler:
            print("default handler:", self.spec.default_handler)
        if self.spec.entry_points:
            print("entry points:")
            for name, entry in self.spec.entry_points.items():
                print(f"  {name}: {entry.get('doc', '')}")
                params = entry.get("parameters")
                if params:
                    for p in params:
                        line = p["name"]
                        if "type" in p:
                            line += f"({p['type']})"
                        line += "  - " + p.get("doc", "")
                        if "default" in p:
                            line += f", default={p['default']}"
                        print("    " + line)

    def skip_image_enrichment(self):
        return False



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

import mlrun.errors


class ImageBuilder(pydantic.v1.BaseModel):
    functionSourceCode: typing.Optional[str] = None  # noqa: N815
    codeEntryType: typing.Optional[str] = None  # noqa: N815
    codeEntryAttributes: typing.Optional[str] = None  # noqa: N815
    source: typing.Optional[str] = None
    code_origin: typing.Optional[str] = None
    origin_filename: typing.Optional[str] = None
    image: typing.Optional[str] = None
    base_image: typing.Optional[str] = None
    commands: typing.Optional[list] = None
    extra: typing.Optional[str] = None
    extra_args: typing.Optional[dict] = None
    builder_env: typing.Optional[dict] = None
    secret: typing.Optional[str] = None
    registry: typing.Optional[str] = None
    load_source_on_run: typing.Optional[bool] = None
    with_mlrun: typing.Optional[bool] = None
    auto_build: typing.Optional[bool] = None
    build_pod: typing.Optional[str] = None
    requirements: typing.Optional[list] = None
    source_code_target_dir: typing.Optional[str] = None

    class Config:
        extra = pydantic.v1.Extra.allow


class LabelsModel(pydantic.v1.BaseModel):
    """
    This class accepts either a dictionary, a list, or a string for filtering by labels.

    :param labels:
        - If a dictionary is provided, it should be in the format {'label_name': 'value'}.
          The values can also be `None`, which will result in the format 'label_name' (without a value).
          This will be converted to a list of strings in the format 'label_name=value'.
        - If a list is provided, all items must be strings. Each string can either
          be a simple label name (e.g., 'label1') or a key-value pair in the format
          'label=value'.
        - If a string is provided, it should be a comma-separated list of labels
          (e.g., 'label1,label2').
        - If no labels are specified, the default is an empty list.
    """

    labels: typing.Optional[
        typing.Union[str, dict[str, typing.Optional[str]], list[str]]
    ]

    @pydantic.v1.validator("labels")
    @classmethod
    def validate(cls, labels) -> list[str]:
        if labels is None:
            return []

        # If labels is a string, split it by commas
        if isinstance(labels, str):
            return [label.strip() for label in labels.split(",") if label.strip()]

        if isinstance(labels, list):
            if not all(isinstance(item, str) for item in labels):
                raise mlrun.errors.MLRunValueError(
                    "All items in the list must be strings."
                )
            return labels

        if isinstance(labels, dict):
            return [
                f"{key}={value}" if value is not None else key
                for key, value in labels.items()
            ]

        raise mlrun.errors.MLRunValueError(
            "Invalid labels format. Must be a string, dictionary of strings, or a list of strings."
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import Counter

import pandas as pd
import semver


def _to_pandas(spark_df):
    """
    Modified version of spark DataFrame.toPandas() -
    https://github.com/apache/spark/blob/v3.2.3/python/pyspark/sql/pandas/conversion.py#L35

    The original code (which is only replaced in pyspark 3.5.0) fails with Pandas 2 installed, with the following error:
    Casting to unit-less dtype 'datetime64' is not supported. Pass e.g. 'datetime64[ns]' instead.

    This modification adds the missing unit to the dtype.
    """
    from pyspark.sql.dataframe import DataFrame
    from pyspark.sql.types import (
        BooleanType,
        IntegralType,
        MapType,
        TimestampType,
    )

    assert isinstance(spark_df, DataFrame)

    from pyspark.sql.pandas.utils import require_minimum_pandas_version

    require_minimum_pandas_version()

    import numpy as np

    timezone = spark_df.sql_ctx._conf.sessionLocalTimeZone()

    if spark_df.sql_ctx._conf.arrowPySparkEnabled():
        use_arrow = True
        try:
            from pyspark.sql.pandas.types import to_arrow_schema
            from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

            require_minimum_pyarrow_version()
            to_arrow_schema(spark_df.schema)
        except Exception as e:
            if spark_df.sql_ctx._conf.arrowPySparkFallbackEnabled():
                msg = (
                    "toPandas attempted Arrow optimization because "
                    "'spark.sql.execution.arrow.pyspark.enabled' is set to true; however, "
                    f"failed by the reason below:\n  {e}\n"
                    "Attempting non-optimization as "
                    "'spark.sql.execution.arrow.pyspark.fallback.enabled' is set to "
                    "true."
                )
                warnings.warn(msg)
                use_arrow = False
            else:
                msg = (
                    "toPandas attempted Arrow optimization because "
                    "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                    "reached the error below and will not continue because automatic fallback "
                    "with 'spark.sql.execution.arrow.pyspark.fallback.enabled' has been set to "
                    f"false.\n  {e}"
                )
                warnings.warn(msg)
                raise

        # Try to use Arrow optimization when the schema is supported and the required version
        # of PyArrow is found, if 'spark.sql.execution.arrow.pyspark.enabled' is enabled.
        if use_arrow:
            try:
                import pyarrow
                from pyspark.sql.pandas.types import (
                    _check_series_localize_timestamps,
                    _convert_map_items_to_dict,
                )

                # Rename columns to avoid duplicated column names.
                tmp_column_names = [f"col_{i}" for i in range(len(spark_df.columns))]
                self_destruct = spark_df.sql_ctx._conf.arrowPySparkSelfDestructEnabled()
                batches = spark_df.toDF(*tmp_column_names)._collect_as_arrow(
                    split_batches=self_destruct
                )
                if len(batches) > 0:
                    table = pyarrow.Table.from_batches(batches)
                    # Ensure only the table has a reference to the batches, so that
                    # self_destruct (if enabled) is effective
                    del batches
                    # Pandas DataFrame created from PyArrow uses datetime64[ns] for date type
                    # values, but we should use datetime.date to match the behavior with when
                    # Arrow optimization is disabled.
                    pandas_options = {"date_as_object": True}
                    if self_destruct:
                        # Configure PyArrow to use as little memory as possible:
                        # self_destruct - free columns as they are converted
                        # split_blocks - create a separate Pandas block for each column
                        # use_threads - convert one column at a time
                        pandas_options.update(
                            {
                                "self_destruct": True,
                                "split_blocks": True,
                                "use_threads": False,
                            }
                        )
                    pdf = table.to_pandas(**pandas_options)
                    # Rename back to the original column names.
                    pdf.columns = spark_df.columns
                    for field in spark_df.schema:
                        if isinstance(field.dataType, TimestampType):
                            pdf[field.name] = _check_series_localize_timestamps(
                                pdf[field.name], timezone
                            )
                        elif isinstance(field.dataType, MapType):
                            pdf[field.name] = _convert_map_items_to_dict(
                                pdf[field.name]
                            )
                    return pdf
                else:
                    return pd.DataFrame.from_records([], columns=spark_df.columns)
            except Exception as e:
                # We might have to allow fallback here as well but multiple Spark jobs can
                # be executed. So, simply fail in this case for now.
                msg = (
                    "toPandas attempted Arrow optimization because "
                    "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                    "reached the error below and can not continue. Note that "
                    "'spark.sql.execution.arrow.pyspark.fallback.enabled' does not have an "
                    "effect on failures in the middle of "
                    f"computation.\n  {e}"
                )
                warnings.warn(msg)
                raise

    # Below is toPandas without Arrow optimization.
    pdf = pd.DataFrame.from_records(spark_df.collect(), columns=spark_df.columns)
    column_counter = Counter(spark_df.columns)

    dtype = [None] * len(spark_df.schema)
    for field_idx, field in enumerate(spark_df.schema):
        # For duplicate column name, we use `iloc` to access it.
        if column_counter[field.name] > 1:
            pandas_col = pdf.iloc[:, field_idx]
        else:
            pandas_col = pdf[field.name]

        pandas_type = _to_corrected_pandas_type(field.dataType)
        # SPARK-21766: if an integer field is nullable and has null values, it can be
        # inferred by pandas as float column. Once we convert the column with NaN back
        # to integer type e.g., np.int16, we will hit exception. So we use the inferred
        # float type, not the corrected type from the schema in this case.
        if pandas_type is not None and not (
            isinstance(field.dataType, IntegralType)
            and field.nullable
            and pandas_col.isnull().any()
        ):
            dtype[field_idx] = pandas_type
        # Ensure we fall back to nullable numpy types, even when whole column is null:
        if isinstance(field.dataType, IntegralType) and pandas_col.isnull().any():
            dtype[field_idx] = np.float64
        if isinstance(field.dataType, BooleanType) and pandas_col.isnull().any():
            dtype[field_idx] = object

    df = pd.DataFrame()
    for index, t in enumerate(dtype):
        column_name = spark_df.schema[index].name

        # For duplicate column name, we use `iloc` to access it.
        if column_counter[column_name] > 1:
            series = pdf.iloc[:, index]
        else:
            series = pdf[column_name]

        if t is not None:
            series = series.astype(t, copy=False)

        # `insert` API makes copy of data, we only do it for Series of duplicate column names.
        # `pdf.iloc[:, index] = pdf.iloc[:, index]...` doesn't always work because `iloc` could
        # return a view or a copy depending by context.
        if column_counter[column_name] > 1:
            df.insert(index, column_name, series, allow_duplicates=True)
        else:
            df[column_name] = series

    pdf = df

    if timezone is None:
        return pdf
    else:
        from pyspark.sql.pandas.types import _check_series_convert_timestamps_local_tz

        for field in spark_df.schema:
            # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
            if isinstance(field.dataType, TimestampType):
                pdf[field.name] = _check_series_convert_timestamps_local_tz(
                    pdf[field.name], timezone
                )
        return pdf


def _to_corrected_pandas_type(dt):
    import numpy as np
    from pyspark.sql.types import (
        BooleanType,
        ByteType,
        DoubleType,
        FloatType,
        IntegerType,
        LongType,
        ShortType,
        TimestampType,
    )

    if isinstance(dt, ByteType):
        return np.int8
    elif isinstance(dt, ShortType):
        return np.int16
    elif isinstance(dt, IntegerType):
        return np.int32
    elif isinstance(dt, LongType):
        return np.int64
    elif isinstance(dt, FloatType):
        return np.float32
    elif isinstance(dt, DoubleType):
        return np.float64
    elif isinstance(dt, BooleanType):
        return bool
    elif isinstance(dt, TimestampType):
        return "datetime64[ns]"
    else:
        return None


def spark_df_to_pandas(spark_df):
    import pyspark

    if semver.parse(pyspark.__version__) >= semver.Version(3, 5, 0):

        def to_pandas(spark_df_inner):
            return spark_df_inner.toPandas()
    else:
        to_pandas = _to_pandas

    # as of pyspark 3.2.3, toPandas fails to convert timestamps unless we work around the issue
    # when we upgrade pyspark, we should check whether this workaround is still necessary
    # see https://stackoverflow.com/questions/76389694/transforming-pyspark-to-pandas-dataframe
    if semver.parse(pd.__version__)["major"] >= 2:
        import pyspark.sql.functions as pyspark_functions

        type_conversion_dict = {}
        for field in spark_df.schema.fields:
            if str(field.dataType) == "TimestampType":
                spark_df = spark_df.withColumn(
                    field.name,
                    pyspark_functions.date_format(
                        pyspark_functions.to_timestamp(field.name),
                        "yyyy-MM-dd'T'HH:mm:ss.SSSSSSSSS",
                    ),
                )
                type_conversion_dict[field.name] = "datetime64[ns]"

        df = to_pandas(spark_df)
        if type_conversion_dict:
            df = df.astype(type_conversion_dict)
        return df
    else:
        return to_pandas(spark_df)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import typing

import mlrun.common.schemas
import mlrun.errors
import mlrun.lists

from .base import NotificationBase, TimedHTTPClient


class GitNotification(NotificationBase):
    """
    API/Client notification for setting a rich run statuses git issue comment (github/gitlab)
    """

    @classmethod
    def validate_params(cls, params):
        git_repo = params.get("repo", None)
        git_issue = params.get("issue", None)
        git_merge_request = params.get("merge_request", None)
        token = (
            params.get("token", None)
            or params.get("GIT_TOKEN", None)
            or params.get("GITHUB_TOKEN", None)
        )
        if not git_repo:
            raise ValueError("Parameter 'repo' is required for GitNotification")

        if not token:
            raise ValueError("Parameter 'token' is required for GitNotification")

        if not git_issue and not git_merge_request:
            raise ValueError(
                "At least one of 'issue' or 'merge_request' is required for GitNotification"
            )

    async def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        git_repo = self.params.get("repo", None)
        git_issue = self.params.get("issue", None)
        git_merge_request = self.params.get("merge_request", None)
        token = (
            self.params.get("token", None)
            or self.params.get("GIT_TOKEN", None)
            or self.params.get("GITHUB_TOKEN", None)
        )
        server = self.params.get("server", None)
        gitlab = self.params.get("gitlab", False)
        await self._pr_comment(
            self._get_html(message, severity, runs, custom_html, alert, event_data),
            git_repo,
            git_issue,
            merge_request=git_merge_request,
            token=token,
            server=server,
            gitlab=gitlab,
        )

    @staticmethod
    async def _pr_comment(
        message: str,
        repo: typing.Optional[str] = None,
        issue: typing.Optional[int] = None,
        merge_request: typing.Optional[int] = None,
        token: typing.Optional[str] = None,
        server: typing.Optional[str] = None,
        gitlab: bool = False,
    ) -> str:
        """push comment message to Git system PR/issue

        :param message:  test message
        :param repo:     repo name (org/repo)
        :param issue:    pull-request/issue number
        :param token:    git system security token
        :param server:   url of the git system
        :param gitlab:   set to True for GitLab (MLRun will try to auto detect the Git system)
        :return:         pr comment id
        """
        if ("CI_PROJECT_ID" in os.environ) or (server and "gitlab" in server):
            gitlab = True
        token = (
            token
            or mlrun.get_secret_or_env("GITHUB_TOKEN")
            or mlrun.get_secret_or_env("GIT_TOKEN")
        )

        if gitlab:
            server = server or "gitlab.com"
            headers = {"PRIVATE-TOKEN": token}
            repo = repo or os.environ.get("CI_PROJECT_ID")
            # auto detect GitLab pr id from the environment
            issue = issue or os.environ.get("CI_ISSUE_IID")
            merge_request = merge_request or os.environ.get("CI_MERGE_REQUEST_IID")
            # replace slash with url encoded slash for GitLab to accept a repo name with slash
            repo = repo.replace("/", "%2F")

            if merge_request:
                url = f"https://{server}/api/v4/projects/{repo}/merge_requests/{merge_request}/notes"
            elif issue:
                url = f"https://{server}/api/v4/projects/{repo}/issues/{issue}/notes"
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "GitLab issue or merge request id not specified"
                )
        else:
            server = server or "api.github.com"
            repo = repo or os.environ.get("GITHUB_REPOSITORY")
            # auto detect pr number if not specified, in github the pr id is identified as an issue id
            # we try and read the pr (issue) id from the github actions event file/object
            if not issue and "GITHUB_EVENT_PATH" in os.environ:
                with open(os.environ["GITHUB_EVENT_PATH"]) as fp:
                    data = fp.read()
                    event = json.loads(data)
                    if "number" not in event:
                        raise mlrun.errors.MLRunInvalidArgumentError(
                            f"issue not found in github actions event\ndata={data}"
                        )
                    issue = event["number"]
            headers = {
                "Accept": "application/vnd.github.v3+json",
                "Authorization": f"token {token}",
            }
            url = f"https://{server}/repos/{repo}/issues/{issue}/comments"

        async with TimedHTTPClient().session() as session:
            resp = await session.post(url, headers=headers, json={"body": message})
            if not resp.ok:
                resp_text = await resp.text()
                raise mlrun.errors.MLRunBadRequestError(
                    f"Failed commenting on PR: {resp_text}"
                )
            data = await resp.json()
            return data.get("id")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ast
import datetime
import os
import random
import sys
import time
from collections import Counter
from copy import copy
from typing import Any, Optional, Union
from urllib.parse import urlparse

import pandas as pd
from mergedeep import merge

import mlrun
import mlrun.utils.helpers
from mlrun.config import config
from mlrun.datastore.snowflake_utils import (
    get_snowflake_password,
    get_snowflake_spark_options,
)
from mlrun.datastore.utils import transform_list_filters_to_tuple
from mlrun.model import DataSource, DataTarget, DataTargetBase, TargetPathObject
from mlrun.utils import logger, now_date
from mlrun.utils.helpers import to_parquet
from mlrun.utils.v3io_clients import get_frames_client

from .. import errors
from ..data_types import ValueType, is_spark_dataframe
from ..platforms.iguazio import parse_path, split_path
from .datastore_profile import datastore_profile_read
from .spark_utils import spark_session_update_hadoop_options
from .utils import (
    _generate_sql_query_with_time_filter,
    filter_df_start_end_time,
    select_columns_from_df,
)


class TargetTypes:
    csv = "csv"
    parquet = "parquet"
    nosql = "nosql"
    redisnosql = "redisnosql"
    tsdb = "tsdb"
    stream = "stream"
    kafka = "kafka"
    dataframe = "dataframe"
    custom = "custom"
    sql = "sql"
    snowflake = "snowflake"

    @staticmethod
    def all():
        return [
            TargetTypes.csv,
            TargetTypes.parquet,
            TargetTypes.nosql,
            TargetTypes.redisnosql,
            TargetTypes.tsdb,
            TargetTypes.stream,
            TargetTypes.kafka,
            TargetTypes.dataframe,
            TargetTypes.custom,
            TargetTypes.sql,
            TargetTypes.snowflake,
        ]


def generate_target_run_id():
    return f"{round(time.time() * 1000)}_{random.randint(0, 999)}"


def write_spark_dataframe_with_options(spark_options, df, mode, write_format=None):
    # TODO: Replace with just df.sparkSession when Spark 3.2 support is dropped
    spark_session = getattr(df, "sparkSession", None) or df.sql_ctx.sparkSession
    non_hadoop_spark_options = spark_session_update_hadoop_options(
        spark_session, spark_options
    )
    if write_format:
        df.write.format(write_format).mode(mode).save(**non_hadoop_spark_options)
    else:
        df.write.mode(mode).save(**non_hadoop_spark_options)


def default_target_names():
    targets = mlrun.mlconf.feature_store.default_targets
    return [target.strip() for target in targets.split(",")]


def get_default_targets(offline_only=False):
    """initialize the default feature set targets list"""
    return [
        DataTargetBase(target, name=str(target), partitioned=(target == "parquet"))
        for target in default_target_names()
        if not offline_only or not target == "nosql"
    ]


def update_targets_run_id_for_ingest(overwrite, targets, targets_in_status):
    run_id = generate_target_run_id()
    for target in targets:
        if overwrite or target.name not in targets_in_status.keys():
            target.run_id = run_id
        else:
            target.run_id = targets_in_status[target.name].run_id


def get_default_prefix_for_target(kind):
    data_prefixes = mlrun.mlconf.feature_store.data_prefixes
    data_prefix = getattr(data_prefixes, kind, None)
    if not data_prefix:
        data_prefix = data_prefixes.default
    return data_prefix


def get_default_prefix_for_source(kind):
    return get_default_prefix_for_target(kind)


def validate_target_paths_for_engine(
    targets, engine, source: Union[DataSource, pd.DataFrame]
):
    """Validating that target paths are suitable for the required engine.
    validate for single file targets only (parquet and csv).

    spark:
        cannot be a single file path (e.g - ends with .csv or .pq)

    storey:
        if csv - must be a single file.
        if parquet - in case of partitioned it must be a directory,
                     else can be both single file or directory

    pandas:
        if source contains chunksize attribute - path must be a directory
        else if parquet - if partitioned(=True) - path must be a directory
        else - path must be a single file


    :param targets:       list of data target objects
    :param engine:        name of the processing engine (storey, pandas, or spark), defaults to storey
    :param source:        source dataframe or other sources (e.g. parquet source see:
                          :py:class:`~mlrun.datastore.ParquetSource` and other classes in
                          mlrun.datastore with suffix Source)
    """
    for base_target in targets:
        if hasattr(base_target, "kind") and (
            base_target.kind == TargetTypes.parquet
            or base_target.kind == TargetTypes.csv
        ):
            target = get_target_driver(base_target)
            is_single_file = target.is_single_file()
            if engine == "spark" and is_single_file:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"spark CSV/Parquet targets must be directories, got path:'{target.path}'"
                )
            elif engine == "pandas":
                # check if source is DataSource (not DataFrame) and if contains chunk size
                if isinstance(source, DataSource) and source.attributes.get(
                    "chunksize"
                ):
                    if is_single_file:
                        raise mlrun.errors.MLRunInvalidArgumentError(
                            "pandas CSV/Parquet targets must be a directory "
                            f"for a chunked source, got path:'{target.path}'"
                        )
                elif target.kind == TargetTypes.parquet and target.partitioned:
                    if is_single_file:
                        raise mlrun.errors.MLRunInvalidArgumentError(
                            "partitioned Parquet target for pandas engine must be a directory, "
                            f"got path:'{target.path}'"
                        )
                elif not is_single_file:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "When using a non chunked source, "
                        f"pandas CSV/Parquet targets must be a single file, got path:'{target.path}'"
                    )
            elif not engine or engine == "storey":
                if target.kind == TargetTypes.csv and not is_single_file:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"CSV target for storey engine must be a single file, got path:'{target.path}'"
                    )
                elif (
                    target.kind == TargetTypes.parquet
                    and target.partitioned
                    and is_single_file
                ):
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"partitioned Parquet target for storey engine must be a directory, got path:'{target.path}'"
                    )


def validate_target_list(targets):
    """Check that no target overrides another target in the list (name/path)"""

    if not targets:
        return
    targets_by_kind_name = [kind for kind in targets if isinstance(kind, str)]
    no_name_target_types_count = Counter(
        [
            target.kind
            for target in targets
            if hasattr(target, "name") and hasattr(target, "kind") and not target.name
        ]
        + targets_by_kind_name
    )
    target_types_requiring_name = [
        target_type
        for target_type, target_type_count in no_name_target_types_count.items()
        if target_type_count > 1
    ]
    if target_types_requiring_name:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Only one default name per target type is allowed (please "
            f"specify name for {target_types_requiring_name} target)"
        )

    target_names_count = Counter(
        [target.name for target in targets if hasattr(target, "name") and target.name]
    )

    targets_with_same_name = [
        target_name
        for target_name, target_name_count in target_names_count.items()
        if target_name_count > 1
    ]

    if targets_with_same_name:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Each target must have a unique name (more than one target with "
            f"those names found {targets_with_same_name})"
        )

    no_path_target_types_count = Counter(
        [
            target.kind
            for target in targets
            if hasattr(target, "path") and hasattr(target, "kind") and not target.path
        ]
        + targets_by_kind_name
    )
    target_types_requiring_path = [
        target_type
        for target_type, target_type_count in no_path_target_types_count.items()
        if target_type_count > 1
    ]
    if target_types_requiring_path:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Only one default path per target type is allowed (please specify "
            f"path for {target_types_requiring_path} target)"
        )

    target_paths_count = Counter(
        [target.path for target in targets if hasattr(target, "path") and target.path]
    )

    targets_with_same_path = [
        target_path
        for target_path, target_path_count in target_paths_count.items()
        if target_path_count > 1
    ]

    if targets_with_same_path:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Each target must have a unique path (more than one target "
            f"with those names found {targets_with_same_path})"
        )


def validate_target_placement(graph, final_step, targets):
    if final_step or graph.is_empty():
        return True
    for target in targets:
        if not target.after_step:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "writer step location is undetermined due to graph branching"
                ", set the target .after_step attribute or the graph .final_step"
            )


def add_target_steps(graph, resource, targets, to_df=False, final_step=None):
    """add the target steps to the graph"""
    targets = targets or []
    key_columns = resource.spec.entities
    timestamp_key = resource.spec.timestamp_key
    features = resource.spec.features
    table = None

    for target in targets:
        # if fset is in passthrough mode, ingest skips writing the data to offline targets
        if resource.spec.passthrough and kind_to_driver[target.kind].is_offline:
            continue

        driver = get_target_driver(target, resource)
        table = driver.get_table_object() or table
        driver.update_resource_status()
        if target.after_step:
            target.attributes["infer_columns_from_data"] = True
        driver.add_writer_step(
            graph,
            target.after_step or final_step,
            features=features if not target.after_step else None,
            key_columns=key_columns,
            timestamp_key=timestamp_key,
            featureset_status=resource.status,
        )
    if to_df:
        # add dataframe target, will return a dataframe
        driver = DFTarget()

        driver.add_writer_step(
            graph,
            final_step,
            features=features,
            key_columns=key_columns,
            timestamp_key=timestamp_key,
        )

    return table


offline_lookup_order = [TargetTypes.parquet, TargetTypes.csv]
online_lookup_order = [TargetTypes.nosql]


def get_offline_target(featureset, name=None):
    """return an optimal offline feature set target"""
    # todo: take status, start_time and lookup order into account
    offline_targets = [
        target
        for target in featureset.status.targets
        if kind_to_driver[target.kind].is_offline
    ]
    target = None
    if name:
        target = next((t for t in offline_targets if t.name == name), None)
    else:
        for kind in offline_lookup_order:
            target = next((t for t in offline_targets if t.kind == kind), None)
            if target:
                break
        if target is None and offline_targets:
            target = offline_targets[0]

    if target:
        return get_target_driver(target, featureset)
    return None


def get_online_target(resource, name=None):
    """return an optimal online feature set target"""
    # todo: take lookup order into account
    for target in resource.status.targets:
        if name and target.name != name:
            continue
        driver = kind_to_driver[target.kind]
        if driver.is_online:
            return get_target_driver(target, resource)
    return None


def get_target_driver(target_spec, resource=None):
    if isinstance(target_spec, dict):
        target_spec = DataTargetBase.from_dict(target_spec)
    driver_class = kind_to_driver[target_spec.kind]
    return driver_class.from_spec(target_spec, resource)


class BaseStoreTarget(DataTargetBase):
    """base target storage driver, used to materialize feature set/vector data"""

    kind = ""
    is_table = False
    suffix = ""
    is_online = False
    is_offline = False
    support_spark = False
    support_storey = False
    support_pandas = False
    support_append = False

    def __init__(
        self,
        name: str = "",
        path=None,
        attributes: Optional[dict[str, str]] = None,
        after_step=None,
        columns=None,
        partitioned: bool = False,
        key_bucketing_number: Optional[int] = None,
        partition_cols: Optional[list[str]] = None,
        time_partitioning_granularity: Optional[str] = None,
        max_events: Optional[int] = None,
        flush_after_seconds: Optional[int] = None,
        storage_options: Optional[dict[str, str]] = None,
        schema: Optional[dict[str, Any]] = None,
    ):
        super().__init__(
            self.kind,
            name,
            path,
            attributes,
            after_step,
            partitioned,
            key_bucketing_number,
            partition_cols,
            time_partitioning_granularity,
            max_events,
            flush_after_seconds,
            schema=schema,
        )

        self.name = name or self.kind
        self.path = str(path) if path is not None else None
        self.after_step = after_step
        self.attributes = attributes or {}
        self.columns = columns or []
        self.partitioned = partitioned
        self.key_bucketing_number = key_bucketing_number
        self.partition_cols = partition_cols
        self.time_partitioning_granularity = time_partitioning_granularity
        self.max_events = max_events
        self.flush_after_seconds = flush_after_seconds
        self.storage_options = storage_options
        self.schema = schema or {}

        self._target = None
        self._resource = None
        self._secrets = {}

    def _get_credential(self, key, default_value=None):
        return mlrun.get_secret_or_env(
            key,
            secret_provider=self._secrets,
            default=default_value,
        )

    def _get_store_and_path(self):
        store, resolved_store_path, url = mlrun.store_manager.get_or_create_store(
            self.get_target_path()
        )
        return store, resolved_store_path, url

    def _get_column_list(self, features, timestamp_key, key_columns, with_type=False):
        result = []
        if self.columns:
            if with_type:
                columns = set(self.columns)
                for feature in features:
                    if feature.name in columns:
                        result.append((feature.name, feature.value_type))
            else:
                result = self.columns
        elif features:
            if with_type:
                for feature in features:
                    result.append((feature.name, feature.value_type))
            else:
                result = list(features.keys())
            if key_columns:
                for key in reversed(key_columns):
                    if key not in result:
                        if with_type:
                            result.insert(0, (key, ValueType.STRING))
                        else:
                            result.insert(0, key)

        if timestamp_key:
            if with_type:
                result = [(timestamp_key, ValueType.DATETIME)] + result
            else:
                result = [timestamp_key] + result

        return result

    def write_dataframe(
        self,
        df,
        key_column=None,
        timestamp_key=None,
        chunk_id=0,
        **kwargs,
    ) -> Optional[int]:
        if is_spark_dataframe(df):
            options = self.get_spark_options(key_column, timestamp_key)
            options.update(kwargs)
            df = self.prepare_spark_df(df, key_column, timestamp_key, options)
            write_format = options.pop("format", None)
            write_spark_dataframe_with_options(
                options, df, "overwrite", write_format=write_format
            )
        elif hasattr(df, "dask"):
            dask_options = self.get_dask_options()
            store, path_in_store, target_path = self._get_store_and_path()
            storage_options = store.get_storage_options()
            df = df.repartition(partition_size="100MB")
            try:
                if dask_options["format"] == "parquet":
                    df.to_parquet(
                        generate_path_with_chunk(self, chunk_id, target_path),
                        storage_options=storage_options,
                    )
                elif dask_options["format"] == "csv":
                    df.to_csv(
                        generate_path_with_chunk(self, chunk_id, target_path),
                        storage_options=storage_options,
                    )
                else:
                    raise NotImplementedError(
                        "Format for writing dask dataframe should be CSV or Parquet!"
                    )
            except Exception as exc:
                raise RuntimeError("Failed to write Dask Dataframe") from exc
        else:
            store, path_in_store, target_path = self._get_store_and_path()
            target_path = generate_path_with_chunk(self, chunk_id, target_path)
            file_system = store.filesystem
            if (
                file_system.protocol == "file"
                # fsspec 2023.10.0 changed protocol from "file" to ("file", "local")
                or isinstance(file_system.protocol, (tuple, list))
                and "file" in file_system.protocol
            ):
                dir = os.path.dirname(target_path)
                if dir:
                    os.makedirs(dir, exist_ok=True)
            target_df = df
            partition_cols = None  # single parquet file
            if not mlrun.utils.helpers.is_parquet_file(target_path):  # directory
                partition_cols = []
                if timestamp_key and (
                    self.partitioned or self.time_partitioning_granularity
                ):
                    target_df = df.copy(deep=False)
                    time_partitioning_granularity = self.time_partitioning_granularity
                    if not time_partitioning_granularity and self.partitioned:
                        time_partitioning_granularity = (
                            mlrun.utils.helpers.DEFAULT_TIME_PARTITIONING_GRANULARITY
                        )
                    for unit, fmt in [
                        ("year", "%Y"),
                        ("month", "%m"),
                        ("day", "%d"),
                        ("hour", "%H"),
                        ("minute", "%M"),
                    ]:
                        partition_cols.append(unit)
                        target_df[unit] = pd.DatetimeIndex(
                            target_df[timestamp_key]
                        ).format(date_format=fmt)
                        if unit == time_partitioning_granularity:
                            break
                # Partitioning will be performed on timestamp_key and then on self.partition_cols
                # (We might want to give the user control on this order as additional functionality)
                partition_cols += self.partition_cols or []

            storage_options = store.get_storage_options()
            if storage_options and self.storage_options:
                storage_options = merge(storage_options, self.storage_options)
            else:
                storage_options = storage_options or self.storage_options

            self._write_dataframe(
                target_df,
                storage_options,
                target_path,
                partition_cols=partition_cols,
                **kwargs,
            )
            try:
                return file_system.size(target_path)
            except Exception:
                return None

    @staticmethod
    def _write_dataframe(df, storage_options, target_path, partition_cols, **kwargs):
        raise NotImplementedError()

    def set_secrets(self, secrets):
        self._secrets = secrets

    def set_resource(self, resource):
        self._resource = resource

    @classmethod
    def from_spec(cls, spec: DataTargetBase, resource=None):
        """create target driver from target spec or other target driver"""
        driver = cls()
        driver.name = spec.name
        driver.path = spec.path
        driver.attributes = spec.attributes
        driver.schema = spec.schema

        if hasattr(spec, "columns"):
            driver.columns = spec.columns

        if hasattr(spec, "_secrets"):
            driver._secrets = spec._secrets

        driver.partitioned = spec.partitioned

        driver.key_bucketing_number = spec.key_bucketing_number
        driver.partition_cols = spec.partition_cols

        driver.time_partitioning_granularity = spec.time_partitioning_granularity
        driver.max_events = spec.max_events
        driver.flush_after_seconds = spec.flush_after_seconds
        driver.storage_options = spec.storage_options

        driver._resource = resource
        driver.run_id = spec.run_id
        driver.after_step = spec.after_step
        return driver

    def get_table_object(self):
        """get storey Table object"""
        return None

    def get_target_path(self):
        path_object = self._target_path_object
        project_name = self._resource.metadata.project if self._resource else None
        return (
            path_object.get_absolute_path(project_name=project_name)
            if path_object
            else None
        )

    def get_target_templated_path(self):
        path_object = self._target_path_object
        return path_object.get_templated_path() if path_object else None

    @property
    def _target_path_object(self):
        """return the actual/computed target path"""
        is_single_file = hasattr(self, "is_single_file") and self.is_single_file()

        if self._resource and self.path:
            parsed_url = urlparse(self.path)
            # When the URL consists only from scheme and endpoint and no path,
            # make a default path for DS and redis targets.
            # Also ignore KafkaTarget when it uses the ds scheme (no default path for KafkaTarget)
            if (
                not isinstance(self, KafkaTarget)
                and parsed_url.scheme in ["ds", "redis", "rediss"]
                and (not parsed_url.path or parsed_url.path == "/")
            ):
                return TargetPathObject(
                    _get_target_path(
                        self,
                        self._resource,
                        self.run_id is not None,
                        netloc=parsed_url.netloc,
                        scheme=parsed_url.scheme,
                    ),
                    self.run_id,
                    is_single_file,
                )

        return self.get_path() or (
            TargetPathObject(
                _get_target_path(self, self._resource, self.run_id is not None),
                self.run_id,
                is_single_file,
            )
            if self._resource
            else None
        )

    def update_resource_status(self, status="", producer=None, size=None):
        """update the data target status"""
        self._target = self._target or DataTarget(
            self.kind, self.name, self.get_target_templated_path()
        )
        target = self._target
        target.attributes = self.attributes
        target.run_id = self.run_id
        target.status = status or target.status or "created"
        target.updated = now_date().isoformat()
        target.size = size
        target.producer = producer or target.producer
        # Copy partitioning-related fields to the status, since these are needed if reading the actual data that
        # is related to the specific target.
        # TODO - instead of adding more fields to the status targets, we should consider changing any functionality
        #       that depends on "spec-fields" to use a merge between the status and the spec targets. One such place
        #       is the fset.to_dataframe() function.
        target.partitioned = self.partitioned
        target.key_bucketing_number = self.key_bucketing_number
        target.partition_cols = self.partition_cols
        target.time_partitioning_granularity = self.time_partitioning_granularity

        self._resource.status.update_target(target)
        return target

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        if not self.support_storey:
            raise mlrun.errors.MLRunRuntimeError(
                f"{type(self).__name__} does not support storey engine"
            )
        raise NotImplementedError()

    def purge(self):
        """
        Delete the files of the target.

        Do not use this function directly from the sdk. Use FeatureSet.purge_targets.
        """
        store, path_in_store, target_path = self._get_store_and_path()
        if path_in_store not in ["", "/"]:
            store.rm(path_in_store, recursive=True)
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Unable to delete target. Please Use purge_targets from FeatureSet object."
            )

    def as_df(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        """return the target data as dataframe"""
        if not self.support_pandas:
            raise NotImplementedError()
        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        return mlrun.get_dataitem(self.get_target_path()).as_df(
            columns=columns,
            df_module=df_module,
            start_time=start_time,
            end_time=end_time,
            time_column=time_column,
            **kwargs,
        )

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        # options used in spark.read.load(**options)
        if not self.support_spark:
            raise mlrun.errors.MLRunRuntimeError(
                f"{type(self).__name__} does not support spark engine"
            )
        raise NotImplementedError()

    def prepare_spark_df(self, df, key_columns, timestamp_key=None, spark_options=None):
        return df

    def get_dask_options(self):
        raise NotImplementedError()

    @property
    def source_spark_attributes(self) -> dict:
        return {}


class ParquetTarget(BaseStoreTarget):
    """Parquet target storage driver, used to materialize feature set/vector data into parquet files.

    :param name:       optional, target name. By default will be called ParquetTarget
    :param path:       optional, Output path. Can be either a file or directory.
     This parameter is forwarded as-is to pandas.DataFrame.to_parquet().
     Default location v3io:///projects/{project}/FeatureStore/{name}/parquet/
    :param attributes: optional, extra attributes for storey.ParquetTarget
    :param after_step: optional, after what step in the graph to add the target
    :param columns:     optional, which columns from data to write
    :param partitioned: optional, whether to partition the file, False by default,
     if True without passing any other partition field, the data will be partitioned by /year/month/day/hour
    :param key_bucketing_number:      optional, None by default will not partition by key,
     0 will partition by the key as is, any other number X will create X partitions and hash the keys to one of them
    :param partition_cols:     optional, name of columns from the data to partition by
    :param time_partitioning_granularity: optional. the smallest time unit to partition the data by.
     For example "hour" will yield partitions of the format /year/month/day/hour
    :param max_events: optional. Maximum number of events to write at a time.
     All events will be written on flow termination,
     or after flush_after_seconds (if flush_after_seconds is set). Default 10k events
    :param flush_after_seconds: optional. Maximum number of seconds to hold events before they are written.
     All events will be written on flow termination, or after max_events are accumulated (if max_events is set).
     Default 15 minutes
    """

    kind = TargetTypes.parquet
    is_offline = True
    support_spark = True
    support_storey = True
    support_dask = True
    support_pandas = True
    support_append = True

    def __init__(
        self,
        name: str = "",
        path=None,
        attributes: Optional[dict[str, str]] = None,
        after_step=None,
        columns=None,
        partitioned: Optional[bool] = None,
        key_bucketing_number: Optional[int] = None,
        partition_cols: Optional[list[str]] = None,
        time_partitioning_granularity: Optional[str] = None,
        max_events: Optional[int] = 10000,
        flush_after_seconds: Optional[int] = 900,
        storage_options: Optional[dict[str, str]] = None,
    ):
        self.path = path
        if partitioned is None:
            partitioned = not self.is_single_file()

        super().__init__(
            name,
            path,
            attributes,
            after_step,
            columns,
            partitioned,
            key_bucketing_number,
            partition_cols,
            time_partitioning_granularity,
            max_events=max_events,
            flush_after_seconds=flush_after_seconds,
            storage_options=storage_options,
        )

        if (
            time_partitioning_granularity is not None
            and time_partitioning_granularity
            not in mlrun.utils.helpers.LEGAL_TIME_UNITS
        ):
            raise errors.MLRunInvalidArgumentError(
                f"time_partitioning_granularity parameter must be one of "
                f"{','.join(mlrun.utils.helpers.LEGAL_TIME_UNITS)}, "
                f"not {time_partitioning_granularity}."
            )

    @staticmethod
    def _write_dataframe(df, storage_options, target_path, partition_cols, **kwargs):
        # In order to save the DataFrame in parquet format, all of the column names must be strings:
        df.columns = [str(column) for column in df.columns.tolist()]
        to_parquet(
            df,
            target_path,
            partition_cols=partition_cols,
            storage_options=storage_options,
            **kwargs,
        )

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        if self.attributes.get("infer_columns_from_data"):
            column_list = None
        else:
            column_list = self._get_column_list(
                features=features,
                timestamp_key=timestamp_key,
                key_columns=None,
                with_type=True,
            )

        # need to extract types from features as part of column list

        partition_cols = None
        if self.key_bucketing_number is not None:
            partition_cols = [("$key", self.key_bucketing_number)]
        if self.partition_cols:
            partition_cols = partition_cols or []
            partition_cols.extend(self.partition_cols)
        time_partitioning_granularity = self.time_partitioning_granularity
        if self.partitioned and all(
            value is None
            for value in [
                time_partitioning_granularity,
                self.key_bucketing_number,
                self.partition_cols,
            ]
        ):
            time_partitioning_granularity = (
                mlrun.utils.helpers.DEFAULT_TIME_PARTITIONING_GRANULARITY
            )
        if time_partitioning_granularity is not None:
            partition_cols = partition_cols or []
            for time_unit in mlrun.utils.helpers.LEGAL_TIME_UNITS:
                partition_cols.append(f"${time_unit}")
                if time_unit == time_partitioning_granularity:
                    break

        target_path = self.get_target_path()
        if not self.partitioned and not mlrun.utils.helpers.is_parquet_file(
            target_path
        ):
            partition_cols = []

        tuple_key_columns = []
        for key_column in key_columns:
            tuple_key_columns.append((key_column.name, key_column.value_type))

        step = graph.add_step(
            name=self.name or "ParquetTarget",
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.ParquetStoreyTarget",
            path=target_path,
            columns=column_list,
            index_cols=tuple_key_columns,
            partition_cols=partition_cols,
            time_field=timestamp_key,
            max_events=self.max_events,
            flush_after_seconds=self.flush_after_seconds,
            update_last_written=featureset_status.update_last_written_for_target,
            **self.attributes,
        )

        original_to_dict = step.to_dict

        def delete_update_last_written(*arg, **kargs):
            result = original_to_dict(*arg, **kargs)
            result["class_args"].pop("update_last_written", None)
            return result

        # update_last_written is not serializable (ML-5108)
        step.to_dict = delete_update_last_written

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        partition_cols = []
        if timestamp_key:
            time_partitioning_granularity = self.time_partitioning_granularity
            if (
                not time_partitioning_granularity
                and self.partitioned
                and not self.partition_cols
            ):
                time_partitioning_granularity = (
                    mlrun.utils.helpers.DEFAULT_TIME_PARTITIONING_GRANULARITY
                )
            if time_partitioning_granularity:
                for unit in mlrun.utils.helpers.LEGAL_TIME_UNITS:
                    partition_cols.append(unit)
                    if unit == time_partitioning_granularity:
                        break

        store, path, url = self._get_store_and_path()
        spark_options = store.get_spark_options(store.spark_url + path)
        spark_options.update(
            {
                "path": store.spark_url + path,
                "format": "parquet",
            }
        )
        for partition_col in self.partition_cols or []:
            partition_cols.append(partition_col)
        if partition_cols:
            spark_options["partitionBy"] = partition_cols
        return spark_options

    def get_dask_options(self):
        return {"format": "parquet"}

    def as_df(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        """return the target data as dataframe"""
        result = mlrun.get_dataitem(self.get_target_path()).as_df(
            columns=columns,
            df_module=df_module,
            format="parquet",
            start_time=start_time,
            end_time=end_time,
            time_column=time_column,
            additional_filters=transform_list_filters_to_tuple(additional_filters),
            **kwargs,
        )
        if not columns:
            drop_cols = []
            if self.time_partitioning_granularity:
                for col in mlrun.utils.helpers.LEGAL_TIME_UNITS:
                    drop_cols.append(col)
                    if col == self.time_partitioning_granularity:
                        break
            elif (
                self.partitioned
                and not self.partition_cols
                and not self.key_bucketing_number
            ):
                drop_cols = mlrun.utils.helpers.DEFAULT_TIME_PARTITIONS
            if drop_cols:
                # if these columns aren't present for some reason, that's no reason to fail
                result.drop(columns=drop_cols, inplace=True, errors="ignore")
        return result

    def is_single_file(self):
        return mlrun.utils.helpers.is_parquet_file(self.path)

    def prepare_spark_df(self, df, key_columns, timestamp_key=None, spark_options=None):
        # If partitioning by time, add the necessary columns
        if (
            timestamp_key
            and isinstance(spark_options, dict)
            and "partitionBy" in spark_options
        ):
            from pyspark.sql.functions import (
                dayofmonth,
                hour,
                minute,
                month,
                second,
                year,
            )

            time_unit_to_op = {
                "year": year,
                "month": month,
                "day": dayofmonth,
                "hour": hour,
                "minute": minute,
                "second": second,
            }
            timestamp_col = df[timestamp_key]
            for partition in spark_options["partitionBy"]:
                if partition not in df.columns and partition in time_unit_to_op:
                    op = time_unit_to_op[partition]
                    df = df.withColumn(partition, op(timestamp_col))
        return df


class CSVTarget(BaseStoreTarget):
    kind = TargetTypes.csv
    suffix = ".csv"
    is_offline = True
    support_spark = True
    support_storey = True
    support_pandas = True

    @staticmethod
    def _write_dataframe(df, storage_options, target_path, partition_cols, **kwargs):
        # avoid writing the range index unless explicitly specified via kwargs
        if isinstance(df.index, pd.RangeIndex):
            kwargs["index"] = kwargs.get("index", False)
        df.to_csv(target_path, storage_options=storage_options, **kwargs)

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        key_columns = list(key_columns.keys())
        column_list = self._get_column_list(
            features=features, timestamp_key=timestamp_key, key_columns=key_columns
        )
        target_path = self.get_target_path()
        graph.add_step(
            name=self.name or "CSVTarget",
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.CSVStoreyTarget",
            path=target_path,
            columns=column_list,
            header=True,
            index_cols=key_columns,
            **self.attributes,
        )

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        store, path, url = self._get_store_and_path()
        spark_options = store.get_spark_options(store.spark_url + path)
        spark_options.update(
            {
                "path": store.spark_url + path,
                "format": "csv",
                "header": "true",
            }
        )
        return spark_options

    def prepare_spark_df(self, df, key_columns, timestamp_key=None, spark_options=None):
        import pyspark.sql.functions as funcs

        for col_name, col_type in df.dtypes:
            # covers TimestampType and TimestampNTZType, which was added in PySpark 3.4.0
            if col_type.startswith("timestamp"):
                # df.write.csv saves timestamps with millisecond precision, but we want microsecond precision
                # for compatibility with storey.
                df = df.withColumn(
                    col_name, funcs.date_format(col_name, "yyyy-MM-dd HH:mm:ss.SSSSSS")
                )
        return df

    def as_df(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        df = super().as_df(
            columns=columns,
            df_module=df_module,
            entities=entities,
            format="csv",
            start_time=start_time,
            end_time=end_time,
            time_column=time_column,
            **kwargs,
        )
        if entities:
            df.set_index(keys=entities, inplace=True)
        return df

    def is_single_file(self):
        if self.path:
            return self.path.endswith(".csv")
        return True


class SnowflakeTarget(BaseStoreTarget):
    """
    :param attributes: A dictionary of attributes for Snowflake connection; will be overridden by database parameters
                       if they exist.
    :param url: Snowflake hostname, in the format: <account_name>.<region>.snowflakecomputing.com
    :param user: Snowflake user for login
    :param db_schema: Database schema
    :param database: Database name
    :param warehouse: Snowflake warehouse name
    :param table_name: Snowflake table name
    """

    support_spark = True
    support_append = True
    is_offline = True
    kind = TargetTypes.snowflake

    def __init__(
        self,
        name: str = "",
        path=None,
        attributes: Optional[dict[str, str]] = None,
        after_step=None,
        columns=None,
        partitioned: bool = False,
        key_bucketing_number: Optional[int] = None,
        partition_cols: Optional[list[str]] = None,
        time_partitioning_granularity: Optional[str] = None,
        max_events: Optional[int] = None,
        flush_after_seconds: Optional[int] = None,
        storage_options: Optional[dict[str, str]] = None,
        schema: Optional[dict[str, Any]] = None,
        url: Optional[str] = None,
        user: Optional[str] = None,
        db_schema: Optional[str] = None,
        database: Optional[str] = None,
        warehouse: Optional[str] = None,
        table_name: Optional[str] = None,
    ):
        attributes = attributes or {}
        if url:
            attributes["url"] = url
        if user:
            attributes["user"] = user
        if database:
            attributes["database"] = database
        if db_schema:
            attributes["db_schema"] = db_schema
        if warehouse:
            attributes["warehouse"] = warehouse
        if table_name:
            attributes["table"] = table_name

        super().__init__(
            name,
            path,
            attributes,
            after_step,
            list(schema.keys()) if schema else columns,
            partitioned,
            key_bucketing_number,
            partition_cols,
            time_partitioning_granularity,
            max_events=max_events,
            flush_after_seconds=flush_after_seconds,
            storage_options=storage_options,
            schema=schema,
        )

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        spark_options = get_snowflake_spark_options(self.attributes)
        spark_options["dbtable"] = self.attributes.get("table")
        return spark_options

    def purge(self):
        import snowflake.connector

        missing = [
            key
            for key in ["database", "db_schema", "table", "url", "user", "warehouse"]
            if self.attributes.get(key) is None
        ]
        if missing:
            raise mlrun.errors.MLRunRuntimeError(
                f"Can't purge Snowflake target, "
                f"some attributes are missing: {', '.join(missing)}"
            )
        account = self.attributes["url"].replace(".snowflakecomputing.com", "")

        with snowflake.connector.connect(
            account=account,
            user=self.attributes["user"],
            password=get_snowflake_password(),
            warehouse=self.attributes["warehouse"],
        ) as snowflake_connector:
            drop_statement = (
                f"DROP TABLE IF EXISTS {self.attributes['database']}.{self.attributes['db_schema']}"
                f".{self.attributes['table']}"
            )
            snowflake_connector.execute_string(drop_statement)

    def as_df(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        raise mlrun.errors.MLRunRuntimeError(
            f"{type(self).__name__} does not support pandas engine"
        )

    @property
    def source_spark_attributes(self) -> dict:
        keys = ["url", "user", "database", "db_schema", "warehouse"]
        attributes = self.attributes or {}
        snowflake_dict = {key: attributes.get(key) for key in keys}
        table = attributes.get("table")
        snowflake_dict["query"] = f"SELECT * from {table}" if table else None
        return snowflake_dict


class NoSqlBaseTarget(BaseStoreTarget):
    is_table = True
    is_online = True
    support_append = True
    support_storey = True
    writer_step_name = "base_name"

    def __new__(cls, *args, **kwargs):
        if cls is NoSqlBaseTarget:
            raise TypeError(f"only children of '{cls.__name__}' may be instantiated")
        return object.__new__(cls)

    def get_table_object(self):
        raise NotImplementedError()

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        table, column_list = self._get_table_and_columns(features, key_columns)

        graph.add_step(
            name=self.name or self.writer_step_name,
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.NoSqlStoreyTarget",
            columns=column_list,
            table=table,
            **self.attributes,
        )

    def _get_table_and_columns(self, features, key_columns):
        key_columns = list(key_columns.keys())
        table = self._resource.uri
        column_list = self._get_column_list(
            features=features,
            timestamp_key=None,
            key_columns=key_columns,
            with_type=True,
        )
        if not self.columns:
            aggregate_features = (
                [key for key, feature in features.items() if feature.aggregate]
                if features
                else []
            )
            column_list = [
                col for col in column_list if col[0] not in aggregate_features
            ]

        return table, column_list

    def prepare_spark_df(self, df, key_columns, timestamp_key=None, spark_options=None):
        raise NotImplementedError()

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        raise NotImplementedError()

    def get_dask_options(self):
        return {"format": "csv"}

    def write_dataframe(
        self, df, key_column=None, timestamp_key=None, chunk_id=0, **kwargs
    ):
        if is_spark_dataframe(df):
            options = self.get_spark_options(key_column, timestamp_key)
            options.update(kwargs)
            df = self.prepare_spark_df(df)
            write_format = options.pop("format", None)
            write_spark_dataframe_with_options(
                options, df, "overwrite", write_format=write_format
            )
        else:
            # To prevent modification of the original dataframe and make sure
            # that the last event of a key is the one being persisted
            if len(df.index.names) and df.index.names[0] is not None:
                df = df.groupby(df.index.names).last()
            else:
                df = df.copy(deep=False)
            access_key = self._get_credential("V3IO_ACCESS_KEY")

            store, path_in_store, target_path = self._get_store_and_path()
            storage_options = store.get_storage_options()
            access_key = storage_options.get("v3io_access_key", access_key)

            _, path_with_container = parse_path(target_path)
            container, path = split_path(path_with_container)

            frames_client = get_frames_client(
                token=access_key, address=config.v3io_framesd, container=container
            )

            frames_client.write("kv", path, df, index_cols=key_column, **kwargs)


class NoSqlTarget(NoSqlBaseTarget):
    kind = TargetTypes.nosql
    support_spark = True
    writer_step_name = "NoSqlTarget"

    def get_table_object(self):
        from storey import Table, V3ioDriver

        store, path_in_store, target_path = self._get_store_and_path()
        endpoint, uri = parse_path(target_path)
        storage_options = store.get_storage_options()
        access_key = storage_options.get("v3io_access_key")

        return Table(
            uri,
            V3ioDriver(webapi=endpoint or mlrun.mlconf.v3io_api, access_key=access_key),
            flush_interval_secs=mlrun.mlconf.feature_store.flush_interval,
        )

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        store, path_in_store, target_path = self._get_store_and_path()
        storage_options = store.get_storage_options()
        store_access_key = storage_options.get("v3io_access_key")
        env_access_key = self._secrets.get(
            "V3IO_ACCESS_KEY", os.getenv("V3IO_ACCESS_KEY")
        )
        if store_access_key and env_access_key and store_access_key != env_access_key:
            logger.warning(
                "The Spark v3io connector does not support access_key parameterization."
                "Spark will disregard the store-provided key."
            )
        spark_options = {
            "path": store.spark_url + path_in_store,
            "format": "io.iguaz.v3io.spark.sql.kv",
        }
        if isinstance(key_column, list) and len(key_column) >= 1:
            spark_options["key"] = key_column[0]
            if len(key_column) > 2:
                spark_options["sorting-key"] = "_spark_object_name"
            if len(key_column) == 2:
                spark_options["sorting-key"] = key_column[1]
        else:
            spark_options["key"] = key_column
        if not overwrite:
            spark_options["columnUpdate"] = True
        return spark_options

    def prepare_spark_df(self, df, key_columns, timestamp_key=None, spark_options=None):
        from pyspark.sql.functions import col

        spark_udf_directory = os.path.dirname(os.path.abspath(__file__))
        sys.path.append(spark_udf_directory)
        try:
            import spark_udf

            df.rdd.context.addFile(spark_udf.__file__)

            for col_name, col_type in df.dtypes:
                if col_type.startswith("decimal("):
                    # V3IO does not support this level of precision
                    df = df.withColumn(col_name, col(col_name).cast("double"))
            if len(key_columns) > 2:
                return df.withColumn(
                    "_spark_object_name",
                    spark_udf.hash_and_concat_v3io_udf(
                        *[col(c) for c in key_columns[1:]]
                    ),
                )
        finally:
            sys.path.remove(spark_udf_directory)
        return df


class RedisNoSqlTarget(NoSqlBaseTarget):
    kind = TargetTypes.redisnosql
    support_spark = True
    writer_step_name = "RedisNoSqlTarget"

    @staticmethod
    def get_server_endpoint(path):
        endpoint, uri = parse_path(path)
        endpoint = endpoint or mlrun.mlconf.redis.url
        if endpoint.startswith("ds://"):
            datastore_profile = datastore_profile_read(endpoint)
            if not datastore_profile:
                raise ValueError(f"Failed to load datastore profile '{endpoint}'")
            if datastore_profile.type != "redis":
                raise ValueError(
                    f"Trying to use profile of type '{datastore_profile.type}' as redis datastore"
                )
            endpoint = datastore_profile.url_with_credentials()
        else:
            parsed_endpoint = urlparse(endpoint)
            if parsed_endpoint.username or parsed_endpoint.password:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Provide Redis username and password only via secrets"
                )
            user = mlrun.get_secret_or_env("REDIS_USER", default="")
            password = mlrun.get_secret_or_env("REDIS_PASSWORD", default="")
            host = parsed_endpoint.hostname
            port = parsed_endpoint.port if parsed_endpoint.port else "6379"
            scheme = parsed_endpoint.scheme
            if user or password:
                endpoint = f"{scheme}://{user}:{password}@{host}:{port}"
            else:
                endpoint = f"{scheme}://{host}:{port}"
        return endpoint, uri

    def get_table_object(self):
        from storey import Table
        from storey.redis_driver import RedisDriver

        endpoint, uri = self.get_server_endpoint(self.get_target_path())

        return Table(
            uri,
            RedisDriver(redis_url=endpoint, key_prefix="/"),
            flush_interval_secs=mlrun.mlconf.feature_store.flush_interval,
        )

    def get_spark_options(self, key_column=None, timestamp_key=None, overwrite=True):
        endpoint, uri = self.get_server_endpoint(self.get_target_path())
        parsed_endpoint = urlparse(endpoint)
        store, path_in_store, path = self._get_store_and_path()
        return {
            "key.column": "_spark_object_name",
            "table": "{" + path_in_store,
            "format": "org.apache.spark.sql.redis",
            "host": parsed_endpoint.hostname,
            "port": parsed_endpoint.port,
            "user": parsed_endpoint.username if parsed_endpoint.username else None,
            "auth": parsed_endpoint.password if parsed_endpoint.password else None,
        }

    def prepare_spark_df(self, df, key_columns, timestamp_key=None, spark_options=None):
        from pyspark.sql.functions import col

        spark_udf_directory = os.path.dirname(os.path.abspath(__file__))
        sys.path.append(spark_udf_directory)
        try:
            import spark_udf

            df.rdd.context.addFile(spark_udf.__file__)

            df = df.withColumn(
                "_spark_object_name",
                spark_udf.hash_and_concat_redis_udf(*[col(c) for c in key_columns]),
            )
        finally:
            sys.path.remove(spark_udf_directory)

        return df

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        table, column_list = self._get_table_and_columns(features, key_columns)

        graph.add_step(
            path=self.get_target_path(),
            name=self.name or self.writer_step_name,
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.RedisNoSqlStoreyTarget",
            columns=column_list,
            table=table,
            **self.attributes,
        )


class StreamTarget(BaseStoreTarget):
    kind = TargetTypes.stream
    is_table = False
    is_online = False
    support_spark = False
    support_storey = True
    support_append = True

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        key_columns = list(key_columns.keys())

        column_list = self._get_column_list(
            features=features, timestamp_key=timestamp_key, key_columns=key_columns
        )
        stream_path = self.get_target_path()
        if not stream_path:
            raise mlrun.errors.MLRunInvalidArgumentError("StreamTarget requires a path")

        graph.add_step(
            name=self.name or "StreamTarget",
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.StreamStoreyTarget",
            columns=column_list,
            stream_path=stream_path,
            **self.attributes,
        )


class KafkaTarget(BaseStoreTarget):
    """
    Kafka target storage driver, used to write data into kafka topics.
    example::
        # define target
        kafka_target = KafkaTarget(
            name="kafka", path="my_topic", brokers="localhost:9092"
        )
        # ingest
        stocks_set.ingest(stocks, [kafka_target])
    :param name:                target name
    :param path:                topic name e.g. "my_topic"
    :param after_step:          optional, after what step in the graph to add the target
    :param columns:             optional, which columns from data to write
    :param producer_options:    additional configurations for kafka producer
    :param brokers:             kafka broker as represented by a host:port pair, or a list of kafka brokers, e.g.
        "localhost:9092", or ["kafka-broker-1:9092", "kafka-broker-2:9092"]
    """

    kind = TargetTypes.kafka
    is_table = False
    is_online = False
    support_spark = False
    support_storey = True
    support_append = True

    def __init__(
        self,
        *args,
        producer_options=None,
        brokers=None,
        **kwargs,
    ):
        attrs = {}

        if brokers:
            attrs["brokers"] = brokers
        if producer_options is not None:
            attrs["producer_options"] = producer_options

        super().__init__(*args, attributes=attrs, **kwargs)

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        key_columns = list(key_columns.keys())
        column_list = self._get_column_list(
            features=features, timestamp_key=timestamp_key, key_columns=key_columns
        )
        path = self.get_target_path()

        if not path:
            raise mlrun.errors.MLRunInvalidArgumentError("KafkaTarget requires a path")

        # Filter attributes to keep only Kafka-related parameters
        # This removes any non-Kafka parameters inherited from BaseStoreTarget
        attributes = mlrun.datastore.utils.KafkaParameters().valid_entries_only(
            self.attributes
        )
        graph.add_step(
            name=self.name or "KafkaTarget",
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.KafkaStoreyTarget",
            columns=column_list,
            path=path,
            attributes=attributes,
        )

    def purge(self):
        pass


class TSDBTarget(BaseStoreTarget):
    kind = TargetTypes.tsdb
    is_table = False
    is_online = False
    support_spark = False
    support_storey = True
    support_append = True

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        key_columns = list(key_columns.keys())
        endpoint, uri = parse_path(self.get_target_path())
        if not timestamp_key:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "feature set timestamp_key must be specified for TSDBTarget writer"
            )

        column_list = self._get_column_list(
            features=features, timestamp_key=None, key_columns=key_columns
        )

        graph.add_step(
            name=self.name or "TSDBTarget",
            class_name="mlrun.datastore.storeytargets.TSDBStoreyTarget",
            after=after,
            graph_shape="cylinder",
            path=uri,
            time_col=timestamp_key,
            index_cols=key_columns,
            columns=column_list,
            **self.attributes,
        )

    def write_dataframe(
        self, df, key_column=None, timestamp_key=None, chunk_id=0, **kwargs
    ):
        access_key = self._secrets.get("V3IO_ACCESS_KEY", os.getenv("V3IO_ACCESS_KEY"))

        new_index = []
        if timestamp_key:
            new_index.append(timestamp_key)
        if key_column:
            if isinstance(key_column, str):
                key_column = [key_column]
            new_index.extend(key_column)

        store, path_in_store, target_path = self._get_store_and_path()
        storage_options = store.get_storage_options()
        access_key = storage_options.get("v3io_access_key", access_key)

        _, path_with_container = parse_path(target_path)
        container, path = split_path(path_with_container)

        frames_client = get_frames_client(
            token=access_key,
            address=config.v3io_framesd,
            container=container,
        )

        frames_client.write(
            "tsdb", path, df, index_cols=new_index if new_index else None, **kwargs
        )


class CustomTarget(BaseStoreTarget):
    kind = "custom"
    is_table = False
    is_online = False
    support_spark = False
    support_storey = True
    support_pandas = True

    def __init__(
        self,
        class_name: str,
        name: str = "",
        after_step=None,
        **attributes,
    ):
        attributes = attributes or {}
        attributes["class_name"] = class_name
        super().__init__(name, "", attributes, after_step=after_step)

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        attributes = copy(self.attributes)
        class_name = attributes.pop("class_name")
        graph.add_step(
            name=self.name,
            after=after,
            graph_shape="cylinder",
            class_name=class_name,
            **attributes,
        )


class DFTarget(BaseStoreTarget):
    kind = TargetTypes.dataframe
    support_storey = True
    support_pandas = True

    def __init__(self, *args, name="dataframe", **kwargs):
        self._df = None
        super().__init__(*args, name=name, **kwargs)

    def set_df(self, df):
        self._df = df

    def update_resource_status(self, status="", producer=None, size=None):
        pass

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        key_columns = list(key_columns.keys())
        # todo: column filter
        graph.add_step(
            name=self.name or "WriteToDataFrame",
            after=after,
            graph_shape="cylinder",
            class_name="storey.ReduceToDataFrame",
            index=key_columns,
            insert_key_column_as=key_columns,
            insert_time_column_as=timestamp_key,
        )

    def as_df(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        return select_columns_from_df(
            filter_df_start_end_time(
                self._df,
                time_column=time_column,
                start_time=start_time,
                end_time=end_time,
            ),
            columns,
        )


class SQLTarget(BaseStoreTarget):
    kind = TargetTypes.sql
    is_online = True
    support_spark = False
    support_storey = True
    support_pandas = True

    def __init__(
        self,
        name: str = "",
        path=None,
        attributes: Optional[dict[str, str]] = None,
        after_step=None,
        partitioned: bool = False,
        key_bucketing_number: Optional[int] = None,
        partition_cols: Optional[list[str]] = None,
        time_partitioning_granularity: Optional[str] = None,
        max_events: Optional[int] = None,
        flush_after_seconds: Optional[int] = None,
        storage_options: Optional[dict[str, str]] = None,
        db_url: Optional[str] = None,
        table_name: Optional[str] = None,
        schema: Optional[dict[str, Any]] = None,
        primary_key_column: str = "",
        if_exists: str = "append",
        create_table: bool = False,
        # create_according_to_data: bool = False,
        varchar_len: int = 50,
        parse_dates: Optional[list[str]] = None,
    ):
        """
        Write to SqlDB as output target for a flow.
        example::
             db_url = "sqlite:///stockmarket.db"
             schema = {'time': datetime.datetime, 'ticker': str,
                    'bid': float, 'ask': float, 'ind': int}
             target = SqlDBTarget(table_name=f'{name}-target', db_url=db_url, create_table=True,
                                   schema=schema, primary_key_column=key)
        :param name:
        :param path:
        :param attributes:
        :param after_step:
        :param partitioned:
        :param key_bucketing_number:
        :param partition_cols:
        :param time_partitioning_granularity:
        :param max_events:
        :param flush_after_seconds:
        :param storage_options:
        :param db_url:                     url string connection to sql database.
                                            If not set, the MLRUN_SQL__URL environment variable will
                                            be used.
        :param table_name:                  the name of the table to access,
                                            from the current database
        :param schema:                      the schema of the table (must pass when
                                            create_table=True)
        :param primary_key_column:          the primary key of the table (must pass always)
        :param if_exists:                   {'fail', 'replace', 'append'}, default 'append'
                                            - fail: If table exists, do nothing.
                                            - replace: If table exists, drop it, recreate it, and insert data.
                                            - append: If table exists, insert data. Create if does not exist.
        :param create_table:                pass True if you want to create new table named by
                                            table_name with schema on current database.
        :param create_according_to_data:    (not valid)
        :param varchar_len :    the defalut len of the all the varchar column (using if needed to create the table).
        :param parse_dates :    all the field to be parsed as timestamp.
        """

        create_according_to_data = False  # TODO: open for user
        db_url = db_url or mlrun.mlconf.sql.url
        if db_url is None or table_name is None:
            attr = {}
        else:
            # check for table existence and acts according to the user input
            self._primary_key_column = primary_key_column

            attr = {
                "table_name": table_name,
                "db_path": db_url,
                "create_according_to_data": create_according_to_data,
                "if_exists": if_exists,
                "parse_dates": parse_dates,
                "varchar_len": varchar_len,
            }
            path = (
                f"mlrunSql://@{db_url}//@{table_name}"
                f"//@{str(create_according_to_data)}//@{if_exists}//@{primary_key_column}//@{create_table}"
            )

        if attributes:
            attributes.update(attr)
        else:
            attributes = attr

        super().__init__(
            name,
            path,
            attributes,
            after_step,
            list(schema.keys()) if schema else None,
            partitioned,
            key_bucketing_number,
            partition_cols,
            time_partitioning_granularity,
            max_events=max_events,
            flush_after_seconds=flush_after_seconds,
            storage_options=storage_options,
            schema=schema,
        )

    def get_table_object(self):
        from storey import SQLDriver, Table

        (db_path, table_name, _, _, primary_key, _) = self._parse_url()
        try:
            primary_key = ast.literal_eval(primary_key)
        except Exception:
            pass
        return Table(
            f"{db_path}/{table_name}",
            SQLDriver(db_path=db_path, primary_key=primary_key),
            flush_interval_secs=mlrun.mlconf.feature_store.flush_interval,
        )

    def add_writer_step(
        self,
        graph,
        after,
        features,
        key_columns=None,
        timestamp_key=None,
        featureset_status=None,
    ):
        key_columns = list(key_columns.keys())
        column_list = self._get_column_list(
            features=features, timestamp_key=timestamp_key, key_columns=key_columns
        )
        table = self._resource.uri
        self._create_sql_table()
        for step in graph.steps.values():
            if step.class_name == "storey.AggregateByKey":
                raise mlrun.errors.MLRunRuntimeError(
                    "SQLTarget does not support aggregation step"
                )
        graph.add_step(
            name=self.name or "SqlTarget",
            after=after,
            graph_shape="cylinder",
            class_name="mlrun.datastore.storeytargets.NoSqlStoreyTarget",
            columns=column_list,
            header=True,
            table=table,
            index_cols=key_columns,
            **self.attributes,
        )

    def as_df(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        try:
            import sqlalchemy

        except (ModuleNotFoundError, ImportError) as exc:
            self._raise_sqlalchemy_import_error(exc)

        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )

        db_path, table_name, _, _, _, _ = self._parse_url()
        engine = sqlalchemy.create_engine(db_path)
        parse_dates: Optional[list[str]] = self.attributes.get("parse_dates")
        with engine.connect() as conn:
            query, parse_dates = _generate_sql_query_with_time_filter(
                table_name=table_name,
                engine=engine,
                time_column=time_column,
                parse_dates=parse_dates,
                start_time=start_time,
                end_time=end_time,
            )
            df = pd.read_sql(
                query,
                con=conn,
                parse_dates=parse_dates,
                columns=columns,
            )
            if self._primary_key_column:
                df.set_index(self._primary_key_column, inplace=True)
        return df

    def write_dataframe(
        self, df, key_column=None, timestamp_key=None, chunk_id=0, **kwargs
    ):
        try:
            import sqlalchemy

        except (ModuleNotFoundError, ImportError) as exc:
            self._raise_sqlalchemy_import_error(exc)

        self._create_sql_table()

        if is_spark_dataframe(df):
            raise ValueError("Spark is not supported")
        else:
            (
                db_path,
                table_name,
                create_according_to_data,
                if_exists,
                primary_key,
                _,
            ) = self._parse_url()
            create_according_to_data = bool(create_according_to_data)
            engine = sqlalchemy.create_engine(
                db_path,
            )
            connection = engine.connect()
            if create_according_to_data:
                # todo : create according to first row.
                pass
            df.to_sql(table_name, connection, if_exists=if_exists)

    def _parse_url(self):
        path = self.path[len("mlrunSql:///") :]
        return path.split("//@")

    def purge(self):
        pass

    def _create_sql_table(self):
        (
            db_path,
            table_name,
            create_according_to_data,
            if_exists,
            primary_key,
            create_table,
        ) = self._parse_url()
        try:
            import sqlalchemy

        except (ModuleNotFoundError, ImportError) as exc:
            self._raise_sqlalchemy_import_error(exc)

        try:
            primary_key = ast.literal_eval(primary_key)
            primary_key_for_check = primary_key
        except Exception:
            primary_key_for_check = [primary_key]
        engine = sqlalchemy.create_engine(db_path)
        with engine.connect() as conn:
            metadata = sqlalchemy.MetaData()
            table_exists = engine.dialect.has_table(conn, table_name)
            if not table_exists and not create_table:
                raise ValueError(f"Table named {table_name} is not exist")

            elif not table_exists and create_table:
                type_to_sql_type = {
                    int: sqlalchemy.Integer,
                    str: sqlalchemy.String(self.attributes.get("varchar_len")),
                    datetime.datetime: sqlalchemy.dialects.mysql.DATETIME(fsp=6),
                    pd.Timestamp: sqlalchemy.dialects.mysql.DATETIME(fsp=6),
                    bool: sqlalchemy.Boolean,
                    float: sqlalchemy.Float,
                    datetime.timedelta: sqlalchemy.Interval,
                    pd.Timedelta: sqlalchemy.Interval,
                }
                # creat new table with the given name
                columns = []
                for col, col_type in self.schema.items():
                    col_type_sql = type_to_sql_type.get(col_type)
                    if col_type_sql is None:
                        raise TypeError(
                            f"'{col_type}' unsupported type for column '{col}'"
                        )
                    columns.append(
                        sqlalchemy.Column(
                            col,
                            col_type_sql,
                            primary_key=(col in primary_key_for_check),
                        )
                    )

                sqlalchemy.Table(table_name, metadata, *columns)
                metadata.create_all(engine)
                if_exists = "append"
                self.path = (
                    f"mlrunSql://@{db_path}//@{table_name}"
                    f"//@{str(create_according_to_data)}//@{if_exists}//@{primary_key}//@{create_table}"
                )
                conn.close()

    @staticmethod
    def _raise_sqlalchemy_import_error(exc):
        raise mlrun.errors.MLRunMissingDependencyError(
            "Using 'SQLTarget' requires sqlalchemy package. Use pip install mlrun[sqlalchemy] to install it."
        ) from exc


kind_to_driver = {
    TargetTypes.parquet: ParquetTarget,
    TargetTypes.csv: CSVTarget,
    TargetTypes.nosql: NoSqlTarget,
    TargetTypes.redisnosql: RedisNoSqlTarget,
    TargetTypes.dataframe: DFTarget,
    TargetTypes.stream: StreamTarget,
    TargetTypes.kafka: KafkaTarget,
    TargetTypes.tsdb: TSDBTarget,
    TargetTypes.custom: CustomTarget,
    TargetTypes.sql: SQLTarget,
    TargetTypes.snowflake: SnowflakeTarget,
}


def _get_target_path(driver, resource, run_id_mode=False, netloc=None, scheme=""):
    """return the default target path given the resource and target kind"""
    kind = driver.kind
    suffix = driver.suffix
    if not suffix:
        if (
            kind == ParquetTarget.kind
            and resource.kind == mlrun.common.schemas.ObjectKind.feature_vector
        ):
            suffix = ".parquet"
    kind_prefix = (
        "sets"
        if resource.kind == mlrun.common.schemas.ObjectKind.feature_set
        else "vectors"
    )
    name = resource.metadata.name
    project = resource.metadata.project or mlrun.mlconf.active_project

    default_kind_name = kind
    if scheme == "ds":
        # "dsnosql" is not an actual target like Parquet or Redis; rather, it serves
        # as a placeholder that can be used in any specified target
        default_kind_name = "dsnosql"
    if scheme == "redis" or scheme == "rediss":
        default_kind_name = TargetTypes.redisnosql

    netloc = netloc or ""
    data_prefix = get_default_prefix_for_target(default_kind_name).format(
        ds_profile_name=netloc,  # In case of ds profile, set its the name
        authority=netloc,  # In case of redis, replace {authority} with netloc
        project=project,
        kind=kind,
        name=name,
    )

    if scheme == "rediss":
        data_prefix = data_prefix.replace("redis://", "rediss://", 1)

    # todo: handle ver tag changes, may need to copy files?
    if not run_id_mode:
        version = resource.metadata.tag
        name = f"{name}-{version or 'latest'}"
    return f"{data_prefix}/{kind_prefix}/{name}{suffix}"


def generate_path_with_chunk(target, chunk_id, path):
    if path is None:
        return ""
    prefix, suffix = os.path.splitext(path)
    if chunk_id and not target.partitioned and not target.time_partitioning_granularity:
        return f"{prefix}/{chunk_id:0>4}{suffix}"
    return path



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import warnings
from typing import Optional
from urllib.parse import urlparse

import boto3
from boto3.s3.transfer import TransferConfig
from fsspec.registry import get_filesystem_class

import mlrun.errors

from .base import DataStore, FileStats, make_datastore_schema_sanitizer


class S3Store(DataStore):
    using_bucket = True

    # TODO: Remove this in 1.12.0
    def _get_endpoint_url_with_deprecation_warning(self):
        """Get S3 endpoint URL with backward compatibility for deprecated S3_ENDPOINT_URL"""
        # First try the new environment variable
        endpoint_url = self._get_secret_or_env("AWS_ENDPOINT_URL_S3")
        if endpoint_url:
            return endpoint_url

        # Check for deprecated environment variable
        deprecated_endpoint_url = self._get_secret_or_env("S3_ENDPOINT_URL")
        if deprecated_endpoint_url:
            warnings.warn(
                "S3_ENDPOINT_URL is deprecated in 1.10.0 and will be removed in 1.12.0, "
                "use AWS_ENDPOINT_URL_S3 instead.",
                # TODO: Remove this in 1.12.0
                FutureWarning,
            )
            return deprecated_endpoint_url

        return None

    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, schema, endpoint, secrets)
        # will be used in case user asks to assume a role and work through fsspec
        self._temp_credentials = None
        region = None

        self.headers = None

        access_key_id = self._get_secret_or_env("AWS_ACCESS_KEY_ID")
        secret_key = self._get_secret_or_env("AWS_SECRET_ACCESS_KEY")
        token_file = self._get_secret_or_env("AWS_CONTAINER_AUTHORIZATION_TOKEN_FILE")
        endpoint_url = self._get_endpoint_url_with_deprecation_warning()
        force_non_anonymous = self._get_secret_or_env("S3_NON_ANONYMOUS")
        profile_name = self._get_secret_or_env("AWS_PROFILE")
        assume_role_arn = self._get_secret_or_env("MLRUN_AWS_ROLE_ARN")

        self.config = TransferConfig(
            multipart_threshold=1024 * 1024 * 25,
            max_concurrency=10,
            multipart_chunksize=1024 * 1024 * 25,
        )

        # If user asks to assume a role, this needs to go through the STS client and retrieve temporary creds
        if assume_role_arn:
            client = boto3.client(
                "sts", aws_access_key_id=access_key_id, aws_secret_access_key=secret_key
            )
            self._temp_credentials = client.assume_role(
                RoleArn=assume_role_arn, RoleSessionName="assumeRoleSession"
            ).get("Credentials")
            if not self._temp_credentials:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"cannot assume role {assume_role_arn}"
                )

            self.s3 = boto3.resource(
                "s3",
                region_name=region,
                aws_access_key_id=self._temp_credentials["AccessKeyId"],
                aws_secret_access_key=self._temp_credentials["SecretAccessKey"],
                aws_session_token=self._temp_credentials["SessionToken"],
                endpoint_url=endpoint_url,
            )
            return

        # User asked for a profile to be used. We don't use access-key or secret-key for this, since the
        # parameters should be in the ~/.aws/credentials file for this to work
        if profile_name:
            session = boto3.session.Session(profile_name=profile_name)
            self.s3 = session.resource(
                "s3",
                region_name=region,
                endpoint_url=endpoint_url,
            )
            return

        if access_key_id or secret_key or force_non_anonymous:
            self.s3 = boto3.resource(
                "s3",
                region_name=region,
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_key,
                endpoint_url=endpoint_url,
            )
        else:
            # from env variables
            self.s3 = boto3.resource(
                "s3", region_name=region, endpoint_url=endpoint_url
            )
            if not token_file:
                # If not using credentials, boto will still attempt to sign the requests, and will fail any operations
                # due to no credentials found. These commands disable signing and allow anonymous mode (same as
                # anon in the storage_options when working with fsspec).
                from botocore.handlers import disable_signing

                self.s3.meta.client.meta.events.register(
                    "choose-signer.s3.*", disable_signing
                )

    @staticmethod
    def get_range(size, offset):
        byterange = f"bytes={offset}-"
        if size:
            byterange += str(offset + size - 1)
        return byterange

    def get_spark_options(self, path=None):
        res = {}
        bucket_str = ""
        if path:
            parsed = urlparse(path)
            if parsed.scheme:  # s3:// or s3a://
                bucket = parsed.hostname
            else:
                # drop a leading slash, if any and take 1st segment
                bucket = path.lstrip("/").split("/", 1)[0]
            bucket_str = f".bucket.{bucket}"

        st = self.get_storage_options()
        if st.get("key"):
            res[f"spark.hadoop.fs.s3a{bucket_str}.access.key"] = st.get("key")
        if st.get("secret"):
            res[f"spark.hadoop.fs.s3a{bucket_str}.secret.key"] = st.get("secret")
        if st.get("endpoint_url"):
            res[f"spark.hadoop.fs.s3a{bucket_str}.endpoint"] = st.get("endpoint_url")
        if st.get("profile"):
            res[f"spark.hadoop.fs.s3a{bucket_str}.aws.profile"] = st.get("profile")
        return res

    @property
    def filesystem(self):
        """return fsspec file system object, if supported"""
        if self._filesystem:
            return self._filesystem
        try:
            import s3fs  # noqa
        except ImportError as exc:
            raise ImportError("AWS s3fs not installed") from exc
        filesystem_class = get_filesystem_class(protocol=self.kind)
        self._filesystem = make_datastore_schema_sanitizer(
            filesystem_class,
            using_bucket=self.using_bucket,
            **self.get_storage_options(),
        )
        return self._filesystem

    def get_storage_options(self):
        force_non_anonymous = self._get_secret_or_env("S3_NON_ANONYMOUS")
        profile = self._get_secret_or_env("AWS_PROFILE")
        endpoint_url = self._get_endpoint_url_with_deprecation_warning()
        access_key_id = self._get_secret_or_env("AWS_ACCESS_KEY_ID")
        secret = self._get_secret_or_env("AWS_SECRET_ACCESS_KEY")
        token_file = self._get_secret_or_env("AWS_CONTAINER_AUTHORIZATION_TOKEN_FILE")

        if self._temp_credentials:
            access_key_id = self._temp_credentials["AccessKeyId"]
            secret = self._temp_credentials["SecretAccessKey"]
            token = self._temp_credentials["SessionToken"]
        else:
            token = None

        storage_options = dict(
            anon=not (force_non_anonymous or (access_key_id and secret) or token_file),
            key=access_key_id,
            secret=secret,
            token=token,
            use_listings_cache=False,
        )

        if endpoint_url:
            client_kwargs = {"endpoint_url": endpoint_url}
            storage_options["client_kwargs"] = client_kwargs

        if profile:
            storage_options["profile"] = profile

        return self._sanitize_options(storage_options)

    @property
    def spark_url(self):
        return f"s3a://{self.endpoint}"

    def get_bucket_and_key(self, key):
        path = self._join(key)[1:]
        return self.endpoint, path

    def upload(self, key, src_path):
        bucket, key = self.get_bucket_and_key(key)
        self.s3.Bucket(bucket).upload_file(src_path, key, Config=self.config)

    def get(self, key, size=None, offset=0):
        bucket, key = self.get_bucket_and_key(key)
        obj = self.s3.Object(bucket, key)
        if size or offset:
            return obj.get(Range=S3Store.get_range(size, offset))["Body"].read()
        return obj.get()["Body"].read()

    def put(self, key, data, append=False):
        data, _ = self._prepare_put_data(data, append)
        bucket, key = self.get_bucket_and_key(key)
        self.s3.Object(bucket, key).put(Body=data)

    def stat(self, key):
        bucket, key = self.get_bucket_and_key(key)
        obj = self.s3.Object(bucket, key)
        size = obj.content_length
        modified = obj.last_modified
        return FileStats(size, time.mktime(modified.timetuple()))

    def listdir(self, key):
        bucket, key = self.get_bucket_and_key(key)
        if not key.endswith("/"):
            key += "/"
        # Object names is S3 are not fully following filesystem semantics - they do not start with /, even for
        # "absolute paths". Therefore, we are removing leading / from path filter.
        if key.startswith("/"):
            key = key[1:]
        key_length = len(key)
        bucket = self.s3.Bucket(bucket)
        return [obj.key[key_length:] for obj in bucket.objects.filter(Prefix=key)]

    def rm(self, path, recursive=False, maxdepth=None):
        bucket, key = self.get_bucket_and_key(path)
        path = f"{bucket}/{key}"
        #  In order to raise an error if there is connection error, ML-7056.
        self.filesystem.exists(path=path)
        self.filesystem.rm(path=path, recursive=recursive, maxdepth=maxdepth)


def parse_s3_bucket_and_key(s3_path):
    try:
        path_parts = s3_path.replace("s3://", "").split("/")
        bucket = path_parts.pop(0)
        key = "/".join(path_parts)
    except Exception as exc:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "failed to parse s3 bucket and key"
        ) from exc

    return bucket, key



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.common.types

from .base import ObjectFormat


class ArtifactFormat(ObjectFormat, mlrun.common.types.StrEnum):
    full = "full"
    minimal = "minimal"

    @staticmethod
    def format_method(_format: str) -> typing.Optional[typing.Callable]:
        return {
            ArtifactFormat.full: None,
            ArtifactFormat.minimal: ArtifactFormat.filter_obj_method(
                [
                    "kind",
                    "metadata",
                    "status.state",
                    "project",
                    "spec.producer",
                    "spec.db_key",
                    "spec.size",
                    "spec.framework",
                    "spec.algorithm",
                    "spec.metrics",
                    "spec.target_path",
                    "spec.parent_uri",
                    "spec.has_children",
                ]
            ),
        }[_format]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tempfile
import urllib.parse
from base64 import b64encode
from copy import copy
from os import path, remove
from typing import Optional, Union
from urllib.parse import urlparse

import fsspec
import orjson
import pandas as pd
import pyarrow
import pytz
import requests

import mlrun.config
import mlrun.errors
from mlrun.datastore.remote_client import BaseRemoteClient
from mlrun.errors import err_to_str
from mlrun.utils import StorePrefix, is_jupyter, logger

from .store_resources import is_store_uri, parse_store_uri
from .utils import filter_df_start_end_time, select_columns_from_df


class FileStats:
    def __init__(self, size, modified, content_type=None):
        self.size = size
        self.modified = modified
        self.content_type = content_type

    def __repr__(self):
        return f"FileStats(size={self.size}, modified={self.modified}, type={self.content_type})"


class DataStore(BaseRemoteClient):
    using_bucket = False

    def __init__(
        self, parent, name, kind, endpoint="", secrets: Optional[dict] = None, **kwargs
    ):
        super().__init__(
            parent=parent, kind=kind, name=name, endpoint=endpoint, secrets=secrets
        )
        self.subpath = ""
        self.options = {}
        self.from_spec = False
        self._filesystem = None

    @property
    def is_structured(self):
        return False

    @property
    def is_unstructured(self):
        return True

    @staticmethod
    def _sanitize_url(url):
        """
        Extract only the schema, netloc, and path from an input URL if they exist,
        excluding parameters, query, or fragments.
        """
        if not url:
            raise mlrun.errors.MLRunInvalidArgumentError("Cannot parse an empty URL")
        parsed_url = urllib.parse.urlparse(url)
        netloc = f"//{parsed_url.netloc}" if parsed_url.netloc else "//"
        return f"{parsed_url.scheme}:{netloc}{parsed_url.path}"

    @staticmethod
    def uri_to_kfp(endpoint, subpath):
        raise ValueError("data store doesnt support KFP URLs")

    @staticmethod
    def uri_to_ipython(endpoint, subpath):
        return ""

    @property
    def filesystem(self) -> Optional[fsspec.AbstractFileSystem]:
        """return fsspec file system object, if supported"""
        return None

    def supports_isdir(self):
        """Whether the data store supports isdir"""
        return True

    def get_storage_options(self):
        """get fsspec storage options"""
        return self._sanitize_options(None)

    def open(self, filepath, mode):
        file_system = self.filesystem
        return file_system.open(filepath, mode)

    def _join(self, key):
        if self.subpath:
            return f"{self.subpath}/{key}"
        return key

    @property
    def spark_url(self):
        return self.url

    def get(self, key, size=None, offset=0):
        pass

    def query(self, key, query="", **kwargs):
        raise ValueError("data store doesnt support structured queries")

    def put(self, key, data, append=False):
        pass

    def _prepare_put_data(self, data, append=False):
        mode = "a" if append else "w"
        if isinstance(data, bytearray):
            data = bytes(data)

        if isinstance(data, bytes):
            return data, f"{mode}b"
        elif isinstance(data, str):
            return data, mode
        else:
            raise TypeError(f"Unable to put a value of type {type(self).__name__}")

    def stat(self, key):
        pass

    def listdir(self, key):
        raise ValueError("data store doesnt support listdir")

    def download(self, key, target_path):
        data = self.get(key)
        mode = "wb"
        if isinstance(data, str):
            mode = "w"
        with open(target_path, mode) as fp:
            fp.write(data)
            fp.close()

    def upload(self, key, src_path):
        pass

    def get_spark_options(self, path=None):
        return {}

    @staticmethod
    def _parquet_reader(
        df_module,
        url,
        file_system,
        time_column,
        start_time,
        end_time,
        additional_filters,
    ):
        from storey.utils import find_filters, find_partitions

        def set_filters(
            partitions_time_attributes,
            start_time_inner,
            end_time_inner,
            filters_inner,
            kwargs,
        ):
            filters = []
            find_filters(
                partitions_time_attributes,
                start_time_inner,
                end_time_inner,
                filters,
                time_column,
            )
            if filters and filters_inner:
                filters[0] += filters_inner

            kwargs["filters"] = filters

        def reader(*args, **kwargs):
            if time_column is None and (start_time or end_time):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "When providing start_time or end_time, must provide time_column"
                )
            if (
                start_time
                and end_time
                and start_time.utcoffset() != end_time.utcoffset()
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "start_time and end_time must have the same time zone"
                )

            if start_time or end_time or additional_filters:
                partitions_time_attributes = find_partitions(url, file_system)
                set_filters(
                    partitions_time_attributes,
                    start_time,
                    end_time,
                    additional_filters,
                    kwargs,
                )
                try:
                    return df_module.read_parquet(*args, **kwargs)
                except pyarrow.lib.ArrowInvalid as ex:
                    if not str(ex).startswith(
                        "Cannot compare timestamp with timezone to timestamp without timezone"
                    ):
                        raise ex

                    start_time_inner = None
                    if start_time:
                        start_time_inner = start_time.replace(
                            tzinfo=None if start_time.tzinfo else pytz.utc
                        )

                    end_time_inner = None
                    if end_time:
                        end_time_inner = end_time.replace(
                            tzinfo=None if end_time.tzinfo else pytz.utc
                        )

                    set_filters(
                        partitions_time_attributes,
                        start_time_inner,
                        end_time_inner,
                        additional_filters,
                        kwargs,
                    )
                    return df_module.read_parquet(*args, **kwargs)
            else:
                return df_module.read_parquet(*args, **kwargs)

        return reader

    def as_df(
        self,
        url,
        subpath,
        columns=None,
        df_module=None,
        format="",
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        df_module = df_module or pd
        file_url = self._sanitize_url(url)
        is_csv, is_json, drop_time_column = False, False, False
        file_system = self.filesystem
        if file_url.endswith(".csv") or format == "csv":
            is_csv = True
            drop_time_column = False
            if columns:
                if (
                    time_column
                    and (start_time or end_time)
                    and time_column not in columns
                ):
                    columns.append(time_column)
                    drop_time_column = True
                kwargs["usecols"] = columns

            reader = df_module.read_csv
            if file_system:
                if file_system.isdir(file_url):

                    def reader(*args, **kwargs):
                        base_path = args[0]
                        file_entries = file_system.listdir(base_path)
                        filenames = []
                        for file_entry in file_entries:
                            if (
                                file_entry["name"].endswith(".csv")
                                and file_entry["size"] > 0
                                and file_entry["type"] == "file"
                            ):
                                filename = file_entry["name"]
                                filename = filename.split("/")[-1]
                                filenames.append(filename)
                        dfs = []
                        if df_module is pd:
                            kwargs.pop("filesystem", None)
                            kwargs.pop("storage_options", None)
                            for filename in filenames:
                                fullpath = f"{base_path}/{filename}"
                                with file_system.open(fullpath) as fhandle:
                                    updated_args = [fhandle]
                                    updated_args.extend(args[1:])
                                    dfs.append(
                                        df_module.read_csv(*updated_args, **kwargs)
                                    )
                        else:
                            for filename in filenames:
                                updated_args = [f"{base_path}/{filename}"]
                                updated_args.extend(args[1:])
                                dfs.append(df_module.read_csv(*updated_args, **kwargs))
                        return df_module.concat(dfs)

        elif mlrun.utils.helpers.is_parquet_file(file_url, format):
            if columns:
                kwargs["columns"] = columns

            reader = self._parquet_reader(
                df_module,
                url,
                file_system,
                time_column,
                start_time,
                end_time,
                additional_filters,
            )

        elif file_url.endswith(".json") or format == "json":
            is_json = True
            reader = df_module.read_json

        else:
            raise Exception(f"File type unhandled {url}")

        if file_system:
            storage_options = self.get_storage_options()
            if url.startswith("ds://"):
                parsed_url = urllib.parse.urlparse(url)
                url = parsed_url.path
                if self.using_bucket:
                    url = url[1:]
                # Pass the underlying file system
                kwargs["filesystem"] = file_system
            elif storage_options:
                kwargs["storage_options"] = storage_options
            df = reader(url, **kwargs)
        else:
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            self.download(self._join(subpath), temp_file.name)
            df = reader(temp_file.name, **kwargs)
            remove(temp_file.name)

        if is_json or is_csv:
            # for parquet file the time filtering is executed in `reader`
            df = filter_df_start_end_time(
                df,
                time_column=time_column,
                start_time=start_time,
                end_time=end_time,
            )
            if drop_time_column:
                df.drop(columns=[time_column], inplace=True)
        if is_json:
            # for csv and parquet files the columns select is executed in `reader`.
            df = select_columns_from_df(df, columns=columns)
        return df

    def to_dict(self):
        return {
            "name": self.name,
            "url": f"{self.kind}://{self.endpoint}/{self.subpath}",
            "secret_pfx": self.secret_pfx,
            "options": self.options,
        }

    def rm(self, path, recursive=False, maxdepth=None):
        try:
            self.filesystem.rm(path=path, recursive=recursive, maxdepth=maxdepth)
        except FileNotFoundError:
            pass

    @staticmethod
    def _is_dd(df_module):
        try:
            import dask.dataframe as dd

            return df_module == dd
        except ImportError:
            return False


class DataItem:
    """Data input/output class abstracting access to various local/remote data sources

    DataItem objects are passed into functions and can be used inside the function, when a function run completes
    users can access the run data via the run.artifact(key) which returns a DataItem object.
    users can also convert a data url (e.g. s3://bucket/key.csv) to a DataItem using `mlrun.get_dataitem(url)`.

    Example::

        # using data item inside a function
        def my_func(context, data: DataItem):
            df = data.as_df()


        # reading run results using DataItem (run.artifact())
        train_run = train_iris_func.run(
            inputs={"dataset": dataset}, params={"label_column": "label"}
        )

        train_run.artifact("confusion-matrix").show()
        test_set = train_run.artifact("test_set").as_df()

        # create and use DataItem from uri
        data = mlrun.get_dataitem("http://xyz/data.json").get()
    """

    def __init__(
        self,
        key: str,
        store: DataStore,
        subpath: str,
        url: str = "",
        meta=None,
        artifact_url=None,
    ):
        self._store = store
        self._key = key
        self._url = url
        self._path = subpath
        self._meta = meta
        self._artifact_url = artifact_url
        self._local_path = ""

    @property
    def key(self):
        """DataItem key"""
        return self._key

    @property
    def suffix(self):
        """DataItem suffix (file extension) e.g. '.png'"""
        _, file_ext = path.splitext(self._path)
        return file_ext

    @property
    def store(self):
        """DataItem store object"""
        return self._store

    @property
    def kind(self):
        """DataItem store kind (file, s3, v3io, ..)"""
        return self._store.kind

    @property
    def meta(self):
        """Artifact Metadata, when the DataItem is read from the artifacts store"""
        return self._meta

    @property
    def artifact_url(self):
        """DataItem artifact url (when its an artifact) or url for simple dataitems"""
        return self._artifact_url or self._url

    @property
    def url(self):
        """DataItem url e.g. /dir/path, s3://bucket/path"""
        return self._url

    def get(
        self,
        size: Optional[int] = None,
        offset: int = 0,
        encoding: Optional[str] = None,
    ) -> Union[bytes, str]:
        """read all or a byte range and return the content

        :param size:     number of bytes to get
        :param offset:   fetch from offset (in bytes)
        :param encoding: encoding (e.g. "utf-8") for converting bytes to str
        :return:         the bytes/str content
        """
        body = self._store.get(self._path, size=size, offset=offset)
        if encoding and isinstance(body, bytes):
            body = body.decode(encoding)
        return body

    def download(self, target_path):
        """download to the target dir/path

        :param target_path: local target path for the downloaded item
        """
        self._store.download(self._path, target_path)

    def put(self, data: Union[bytes, str], append: bool = False) -> None:
        """write/upload the data, append is only supported by some datastores

        :param data:   data (bytes/str) to write
        :param append: append data to the end of the object, NOT SUPPORTED BY SOME OBJECT STORES!
        """
        self._store.put(self._path, data, append=append)

    def delete(self):
        """delete the object from the datastore"""
        self._store.rm(self._path)

    def upload(self, src_path):
        """upload the source file (src_path)

        :param src_path: source file path to read from and upload
        """
        self._store.upload(self._path, src_path)

    def stat(self):
        """return FileStats class (size, modified, content_type)"""
        return self._store.stat(self._path)

    def open(self, mode):
        """return fsspec file handler, if supported"""
        return self._store.open(self._url, mode)

    def ls(self):
        """return a list of child file names"""
        return self._store.listdir(self._path)

    def listdir(self):
        """return a list of child file names"""
        return self._store.listdir(self._path)

    def local(self):
        """get the local path of the file, download to tmp first if it's a remote object"""
        if self.kind == "file":
            return self._path
        if self._local_path:
            return self._local_path

        dot = self._path.rfind(".")
        suffix = "" if dot == -1 else self._path[dot:]
        temp_file = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
        self._local_path = temp_file.name
        logger.info(f"downloading {self.url} to local temp file")
        self.download(self._local_path)
        return self._local_path

    def remove_local(self):
        """remove the local file if it exists and was downloaded from a remote object"""
        if self.kind == "file":
            return

        if self._local_path:
            remove(self._local_path)
            self._local_path = ""

    def as_df(
        self,
        columns=None,
        df_module=None,
        format="",
        time_column=None,
        start_time=None,
        end_time=None,
        additional_filters=None,
        **kwargs,
    ):
        """return a dataframe object (generated from the dataitem).

        :param columns:     optional, list of columns to select
        :param df_module:   optional, py module used to create the DataFrame (e.g. pd, dd, cudf, ..)
        :param format:      file format, if not specified it will be deducted from the suffix
        :param start_time:  filters out data before this time
        :param end_time:    filters out data after this time
        :param time_column: Store timestamp_key will be used if None.
                            The results will be filtered by this column and start_time & end_time.
        :param additional_filters: List of additional_filter conditions as tuples.
                                    Each tuple should be in the format (column_name, operator, value).
                                    Supported operators: "=", ">=", "<=", ">", "<".
                                    Example: [("Product", "=", "Computer")]
                                    For all supported filters, please see:
                                    https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
        """
        df = self._store.as_df(
            self._url,
            self._path,
            columns=columns,
            df_module=df_module,
            format=format,
            time_column=time_column,
            start_time=start_time,
            end_time=end_time,
            additional_filters=additional_filters,
            **kwargs,
        )
        return df

    def show(self, format: Optional[str] = None) -> None:
        """show the data object content in Jupyter

        :param format: format to use (when there is no/wrong suffix), e.g. 'png'
        """
        if not is_jupyter:
            logger.warning(
                "Jupyter was not detected. `.show()` displays only inside Jupyter."
            )
            return

        from IPython import display

        suffix = self.suffix.lower()
        if format:
            suffix = "." + format

        if suffix in [".jpg", ".png", ".gif"]:
            display.display(display.Image(self.get(), format=suffix[1:]))
        elif suffix in [".htm", ".html"]:
            display.display(display.HTML(self.get(encoding="utf-8")))
        elif suffix in [".csv", ".pq", ".parquet"]:
            display.display(self.as_df())
        elif suffix in [".yaml", ".txt", ".py"]:
            display.display(display.Pretty(self.get(encoding="utf-8")))
        elif suffix == ".json":
            display.display(display.JSON(orjson.loads(self.get())))
        elif suffix == ".md":
            display.display(display.Markdown(self.get(encoding="utf-8")))
        else:
            logger.error(f"unsupported show() format {suffix} for {self.url}")

    def get_artifact_type(self) -> Union[str, None]:
        """
        Check if the data item represents an Artifact (one of Artifact, DatasetArtifact and ModelArtifact). If it does
        it return the store uri prefix (artifacts, datasets or models), otherwise None.

        :return: The store prefix of the artifact if it is an artifact data item and None if not.
        """
        if self.artifact_url and is_store_uri(url=self.artifact_url):
            store_uri_prefix = parse_store_uri(self.artifact_url)[0]
            if StorePrefix.is_artifact(prefix=store_uri_prefix):
                return store_uri_prefix
        return None

    def __str__(self):
        return self.url

    def __repr__(self):
        return f"'{self.url}'"


def basic_auth_header(user, password):
    username = user.encode("latin1")
    password = password.encode("latin1")
    base = b64encode(b":".join((username, password))).strip()
    authstr = "Basic " + base.decode("ascii")
    return {"Authorization": authstr}


class HttpStore(DataStore):
    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, schema, endpoint, secrets)
        self._https_auth_token = None
        self._schema = schema
        self.auth = None
        self._headers = {}
        self._enrich_https_token()
        self._validate_https_token()

    @property
    def filesystem(self):
        """return fsspec file system object, if supported"""
        if not self._filesystem:
            self._filesystem = fsspec.filesystem("http")
        return self._filesystem

    def supports_isdir(self):
        return False

    def upload(self, key, src_path):
        raise ValueError("unimplemented")

    def put(self, key, data, append=False):
        raise ValueError("unimplemented")

    def get(self, key, size=None, offset=0):
        headers = self._headers
        if urlparse(self.url).hostname == "api.github.com":
            headers = copy(self._headers)
            headers["Accept"] = headers.get("Accept", "application/vnd.github.raw")
        data = self._http_get(self.url + self._join(key), headers, self.auth)
        if offset:
            data = data[offset:]
        if size:
            data = data[:size]
        return data

    def _enrich_https_token(self):
        token = self._get_secret_or_env("HTTPS_AUTH_TOKEN")
        if token:
            self._https_auth_token = token
            self._headers.setdefault("Authorization", f"Bearer {token}")

    def _validate_https_token(self):
        if self._https_auth_token and self._schema in ["http"]:
            logger.warn(
                f"A AUTH TOKEN should not be provided while using {self._schema} "
                f"schema as it is not secure and is not recommended."
            )

    def _http_get(
        self,
        url,
        headers=None,
        auth=None,
    ):
        # import here to prevent import cycle
        from mlrun.config import config as mlconf

        verify_ssl = mlconf.httpdb.http.verify
        try:
            response = requests.get(url, headers=headers, auth=auth, verify=verify_ssl)
        except OSError as exc:
            raise OSError(f"error: cannot connect to {url}: {err_to_str(exc)}")

        mlrun.errors.raise_for_status(response)
        return response.content


# This wrapper class is designed to extract the 'ds' schema and profile name from URL-formatted paths.
# Within fsspec, the AbstractFileSystem::_strip_protocol() internal method is used to handle complete URL paths.
# As an example, it converts an S3 URL 's3://s3bucket/path' to just 's3bucket/path'.
# Since 'ds' schemas are not inherently processed by fsspec, we have adapted the _strip_protocol()
# method specifically to strip away the 'ds' schema as required.
def make_datastore_schema_sanitizer(cls, using_bucket=False, *args, **kwargs):
    if not issubclass(cls, fsspec.AbstractFileSystem):
        raise ValueError("Class must be a subclass of fsspec.AbstractFileSystem")

    class DatastoreSchemaSanitizer(cls):
        @classmethod
        def _strip_protocol(cls, url):
            if url.startswith("ds://"):
                parsed_url = urlparse(url)
                url = parsed_url.path
                if using_bucket:
                    url = url[1:]
            return super()._strip_protocol(url)

    return DatastoreSchemaSanitizer(*args, **kwargs)
